// Import BenchmarkBase class.
import 'package:benchmark_harness/benchmark_harness.dart';

const double _epsilon = 0.0001;

bool nearEqual(double? a, double? b, double epsilon) {
  assert(epsilon != null);
  assert(epsilon >= 0.0);
  if (a == null || b == null) return a == b;
  return (a > (b - epsilon)) && (a < (b + epsilon)) || a == b;
}

abstract class RectFinder {
  const RectFinder();
  String get title;
  LayoutRect findFirstRectFor(
    int requiredRails,
    Iterable<double> scrollOffsets,
  );
}

// Defines an relative rect were a child can be layout
// In the main axis the top limit is minOffset and bottom limit maxOffset
// In the cross axis the start limit is startIndex and the width is crossAxisCount
class LayoutRect {
  const LayoutRect(
    this.startRailIndex,
    this.rails,
    this.leadingScrollOffset,
    this.trailingScrollOffset,
  );

  // Main axis
  final double leadingScrollOffset;
  final double trailingScrollOffset;

  // Cross axis
  final int startRailIndex;
  final int rails;
}

class LayoutOffset {
  const LayoutOffset(
    this.startIndex,
    this.scrollOffset,
  );
  // Main axis
  final double scrollOffset;
  // Cross axis
  final int startIndex;
}

class Finder1 extends RectFinder {
  const Finder1();

  /// Finds the first available rect for the specified [scrollOffsets] list.
  LayoutRect findFirstAvailableRect(
    List<double> scrollOffsets,
    int requiredRails,
  ) {
    int startIndex = 0;
    int crossAxisCount = 1;
    bool contiguous = false;
    double leadingScrollOffset = double.infinity;
    double trailingScrollOffset = double.infinity;

    // We have to use the nearEqual function because of floating-point arithmetic.
    // Ex: 0.1 + 0.2 = 0.30000000000000004 and not 0.3.
    for (var i = startIndex; i < scrollOffsets.length; ++i) {
      final double offset = scrollOffsets[i];
      final bool equalOffset = nearEqual(offset, leadingScrollOffset, _epsilon);

      if (equalOffset && contiguous) {
        crossAxisCount++;
      } else if (!equalOffset) {
        if (offset < leadingScrollOffset) {
          startIndex = i;
          crossAxisCount = 1;
          contiguous = true;
          trailingScrollOffset = leadingScrollOffset;
          leadingScrollOffset = offset;
        } else if (offset > leadingScrollOffset &&
            offset < trailingScrollOffset) {
          contiguous = false;
          trailingScrollOffset = offset;
        }
      } else {
        contiguous = false;
      }
    }
    return LayoutRect(
      startIndex,
      crossAxisCount,
      leadingScrollOffset,
      trailingScrollOffset,
    );
  }

  /// Finds the first available rect with at least the specified [requiredRails].
  LayoutRect findFirstRectFor(
    int requiredRails,
    Iterable<double> scrollOffsets,
  ) {
    assert(scrollOffsets.length >= requiredRails,
        'Available rails are less than the ones requested');
    // Copy offsets list so it can be modified
    final List<double> offsets = scrollOffsets.toList();
    final LayoutRect rect = findFirstAvailableRect(offsets, requiredRails);
    if (rect.rails < requiredRails) {
      // Not enough space for the specified cross axis count.
      // We have to fill this block and try again.
      for (var i = 0; i < rect.rails; ++i) {
        offsets[i + rect.startRailIndex] = rect.trailingScrollOffset;
      }
      return findFirstRectFor(requiredRails, offsets);
    } else {
      return rect;
    }
  }

  @override
  String get title => '1';
}

class Finder2 extends RectFinder {
  LayoutRect findFirstAvailableRect(
    List<double> scrollOffsets,
    int requiredRails,
  ) {
    LayoutRect? _targetRect;
    int startIndex = 0;
    int crossAxisCount = 1;
    bool contiguous = false;
    double leadingScrollOffset = double.infinity;
    double trailingScrollOffset = double.infinity;

    // We have to use the nearEqual function because of floating-point arithmetic.
    // Ex: 0.1 + 0.2 = 0.30000000000000004 and not 0.3.
    for (var i = startIndex; i < scrollOffsets.length; ++i) {
      final double offset = scrollOffsets[i];
      final bool equalOffset = nearEqual(offset, leadingScrollOffset, _epsilon);

      if (equalOffset && contiguous) {
        crossAxisCount++;
      } else if (!equalOffset) {
        if (offset < leadingScrollOffset) {
          startIndex = i;
          crossAxisCount = 1;
          contiguous = true;
          trailingScrollOffset = leadingScrollOffset;
          leadingScrollOffset = offset;
        } else if (offset > leadingScrollOffset &&
            offset < trailingScrollOffset) {
          contiguous = false;
          trailingScrollOffset = offset;

          if (crossAxisCount >= requiredRails) {
            if (_targetRect == null ||
                _targetRect.leadingScrollOffset > leadingScrollOffset) {
              _targetRect = LayoutRect(
                startIndex,
                crossAxisCount,
                leadingScrollOffset,
                trailingScrollOffset,
              );
            }
          }
        }
      } else {
        contiguous = false;
        if (crossAxisCount >= requiredRails) {
          if (_targetRect == null ||
              _targetRect.leadingScrollOffset > leadingScrollOffset) {
            _targetRect = LayoutRect(
              startIndex,
              crossAxisCount,
              leadingScrollOffset,
              trailingScrollOffset,
            );
          }
        }
      }
    }

    return _targetRect ??
        LayoutRect(
          startIndex,
          crossAxisCount,
          leadingScrollOffset,
          trailingScrollOffset,
        );
  }

  /// Finds the first available rect with at least the specified [requiredRails].
  LayoutRect findFirstRectFor(
    int requiredRails,
    Iterable<double> scrollOffsets,
  ) {
    assert(scrollOffsets.length >= requiredRails,
        'Available rails are less than the ones requested');
    // Copy offsets list so it can be modified
    final List<double> offsets = scrollOffsets.toList();
    final LayoutRect rect = findFirstAvailableRect(offsets, requiredRails);
    if (rect.rails < requiredRails) {
      // Not enough space for the specified cross axis count.
      // We have to fill this block and try again.
      for (var i = 0; i < rect.rails; ++i) {
        offsets[i + rect.startRailIndex] = rect.trailingScrollOffset;
      }
      return findFirstRectFor(requiredRails, offsets);
    } else {
      return rect;
    }
  }

  @override
  String get title => '2';
}

class Finder3 extends RectFinder {
  /// Finds the first available rect with at least the specified [requiredRails].
  LayoutRect findFirstRectFor(int requiredRails, Iterable<double> scrollOffsets,
      {int initialIndex = 0}) {
    assert(scrollOffsets.length >= requiredRails,
        'Available rails are less than the ones requested');
    final List<double> offsets = scrollOffsets.toList();
    LayoutRect? _candidateRect;
    LayoutRect? _possibleRect;
    double maxOffset = 0;
    int startIndex = 0;
    int crossAxisCount = 1;
    bool contiguous = false;
    double leadingScrollOffset = double.infinity;

    // We have to use the nearEqual function because of floating-point arithmetic.
    // Ex: 0.1 + 0.2 = 0.30000000000000004 and not 0.3.
    for (var i = initialIndex; i < scrollOffsets.length; ++i) {
      final double offset = offsets[i];
      final bool equalOffset = nearEqual(offset, leadingScrollOffset, _epsilon);
      if (maxOffset < offset) maxOffset = offset;

      if (equalOffset && contiguous) {
        crossAxisCount++;
      } else {
        if (crossAxisCount >= requiredRails) {
          if (_candidateRect == null ||
              _candidateRect.leadingScrollOffset > leadingScrollOffset) {
            _candidateRect = LayoutRect(
              startIndex,
              crossAxisCount,
              leadingScrollOffset,
              0,
            );
          }
        }
        if (!equalOffset && offset < leadingScrollOffset) {
          if (i != initialIndex &&
              (_possibleRect == null ||
                  _possibleRect.leadingScrollOffset > leadingScrollOffset)) {
            _possibleRect = LayoutRect(
              startIndex,
              crossAxisCount,
              leadingScrollOffset,
              0,
            );
          }
          contiguous = true;
          startIndex = i;
          crossAxisCount = 1;
          leadingScrollOffset = offset;
        } else {
          contiguous = false;
        }
      }
    }

    if (crossAxisCount >= requiredRails) {
      return LayoutRect(
        startIndex,
        crossAxisCount,
        leadingScrollOffset,
        0,
      );
    }

    if (_candidateRect == null) {
      for (var i = 0; i < offsets.length; ++i) {
        offsets[i] = _possibleRect?.leadingScrollOffset ?? maxOffset;
      }
      return findFirstRectFor(
        requiredRails,
        offsets,
        // initialIndex: _possibleRect?.startRailIndex ?? 0,
      );
    } else {
      return _candidateRect;
    }
  }

  @override
  String get title => '3';
}

class Finder4 extends RectFinder {
  /// Finds the first available rect with at least the specified [requiredRails].
  LayoutRect findFirstRectFor(int requiredRails, Iterable<double> scrollOffsets,
      {int initialIndex = 0}) {
    assert(scrollOffsets.length != 0,
        'There should be at least one rail available');
    assert(scrollOffsets.length >= requiredRails,
        'Available rails are less than the ones requested');
    final List<double> offsets = scrollOffsets.toList();
    LayoutOffset? _candidatePosition;
    LayoutOffset? _possiblePosition;

    int startRailIndex = 0;
    int railCount = 1;
    bool contiguous = false;
    double minScrollOffset = double.infinity;
    double maxScrollOffset = 0;

    void setCandidateOffsetIfMeetsRequiremets() {
      if (_candidatePosition == null ||
          _candidatePosition!.scrollOffset > minScrollOffset) {
        _candidatePosition = LayoutOffset(startRailIndex, minScrollOffset);
      }
    }

    void setPossibleOffsetIfMeetsRequirements() {
      if (_possiblePosition == null ||
          _possiblePosition!.scrollOffset > minScrollOffset) {
        _possiblePosition = LayoutOffset(startRailIndex, minScrollOffset);
      }
    }

    // We have to use the nearEqual function because of floating-point arithmetic.
    // Ex: 0.1 + 0.2 = 0.30000000000000004 and not 0.3.
    for (var i = initialIndex; i < scrollOffsets.length; ++i) {
      final double offset = offsets[i];
      final bool equalOffset = nearEqual(offset, minScrollOffset, _epsilon);
      if (maxScrollOffset < offset) maxScrollOffset = offset;

      // This algorithm checks first if the current available rail is contiguous
      // to the previous one and they have the same scrollOffset. In that case
      // the rail count of the available space it is increased by one.
      //
      // In the contrary case, first it check that the previous space is valid
      // to be the candidate.
      //
      // Secondly if the current new rail is positioned with a minor offset than
      // the candidate we set the previous space as possible candidate
      // and we analyze the new space.
      // Otherwise if the new rail is positioned at a mayor offset we set the
      // contagious to false
      if (equalOffset && contiguous) {
        railCount++;
      } else {
        if (contiguous && railCount < requiredRails) {
          setCandidateOffsetIfMeetsRequiremets();
        }
        if (!equalOffset && offset < minScrollOffset) {
          if (i != initialIndex && _candidatePosition == null) {
            setPossibleOffsetIfMeetsRequirements();
          }
          contiguous = true;
          startRailIndex = i;
          railCount = 1;
          minScrollOffset = offset;
        } else {
          contiguous = false;
        }
      }
    }

    if (railCount >= requiredRails) {
      return LayoutRect(
        startRailIndex,
        railCount,
        minScrollOffset,
        0,
      );
    }

    if (_candidatePosition == null) {
      for (var i = 0; i < offsets.length; ++i) {
        offsets[i] = _possiblePosition?.scrollOffset ?? maxScrollOffset;
      }
      return findFirstRectFor(
        requiredRails,
        offsets,
        // initialIndex: _possibleRect?.startRailIndex ?? 0,
      );
    } else {
      return LayoutRect(
        _candidatePosition!.startIndex,
        requiredRails,
        _candidatePosition!.scrollOffset,
        0,
      );
    }
  }

  @override
  String get title => '4';
}

typedef ListBuilder = List<double> Function(int length);

class FindRectBenchmark extends BenchmarkBase {
  FindRectBenchmark([
    this.finder = const Finder1(),
    this.rails = 200,
    this.length = 200,
    this.listBuilder = defaultListBuilder,
  ]) : super('FindRectBenchmark ${finder.title} - $length');

  final int length;
  final int rails;
  final ListBuilder listBuilder;
  final RectFinder finder;

  late List<double> scrollOffsets;
  late LayoutRect rect;

  static List<double> defaultListBuilder(int length) {
    return List.generate(length, (index) => index * 100.0);
  }

  static List<double> reversed(int length) {
    return List.generate(length, (index) => index * 100.0);
  }

  // The benchmark code.
  @override
  void run() {
    rect = finder.findFirstRectFor(12, scrollOffsets);
  }

  // Not measured setup code executed prior to the benchmark runs.
  @override
  void setup() {
    scrollOffsets = listBuilder(length);
  }

  // Not measured teardown code executed after the benchmark runs.
  @override
  void teardown() {
    // print(rect.startRailIndex);
  }
}

List<double> list(int length) {
  return List.generate(length, (index) => index * 100.0);
}

List<double> reversed(int length) {
  return List.generate(length, (index) => index * 100.0).reversed.toList();
}

List<double> pyramid(int length) {
  return List.generate(
    length,
    (index) => (length - index + 0.5).abs() * 100.0,
  ).toList();
}

main() {
  // Run TemplateBenchmark
  //FindRectBenchmark(Finder3(), 12, false).report();
  //FindRectBenchmark(Finder1(), 12, false).report();
  //FindRectBenchmark(Finder1(), 12, true).report();
  //FindRectBenchmark(Finder1(), 100, false).report();
  //FindRectBenchmark(Finder1(), 100, true).report();
  //FindRectBenchmark(Finder1(), 1000, false).report();
  //FindRectBenchmark(Finder1(), 1000, true).report();
  //FindRectBenchmark(Finder1(), 10000, false).report();
  //FindRectBenchmark(Finder1(), 10000, true).report();
  //print('\n');
  //FindRectBenchmark(Finder2(), 12, false).report();
  //FindRectBenchmark(Finder2(), 12, true).report();
  //FindRectBenchmark(Finder2(), 100, false).report();
  //FindRectBenchmark(Finder2(), 100, true).report();
  //FindRectBenchmark(Finder2(), 1000, false).report();
  //FindRectBenchmark(Finder2(), 1000, true).report();
  //FindRectBenchmark(Finder2(), 10000, false).report();
  //FindRectBenchmark(Finder2(), 10000, true).report();
  print('\n');
  print('1');
  FindRectBenchmark(Finder1(), 1, 12, list).report();
  FindRectBenchmark(Finder2(), 1, 12, list).report();
  FindRectBenchmark(Finder3(), 1, 12, list).report();
  FindRectBenchmark(Finder4(), 1, 12, list).report();
  print('\n');
  print('1 - Reversed');
  FindRectBenchmark(Finder1(), 1, 12, reversed).report();
  FindRectBenchmark(Finder2(), 1, 12, reversed).report();
  FindRectBenchmark(Finder3(), 1, 12, reversed).report();
  FindRectBenchmark(Finder4(), 1, 12, reversed).report();
  print('\n');
  print('1 - pyramid');
  FindRectBenchmark(Finder1(), 1, 12, pyramid).report();
  FindRectBenchmark(Finder2(), 1, 12, pyramid).report();
  FindRectBenchmark(Finder3(), 1, 12, pyramid).report();
  FindRectBenchmark(Finder4(), 1, 12, pyramid).report();
  print('\n');
  print('12');
  FindRectBenchmark(Finder1(), 12, 12, list).report();
  FindRectBenchmark(Finder2(), 12, 12, list).report();
  FindRectBenchmark(Finder3(), 12, 12, list).report();
  FindRectBenchmark(Finder4(), 12, 12, list).report();
  print('\n');
  print('12 - reversed');
  FindRectBenchmark(Finder1(), 12, 12, reversed).report();
  FindRectBenchmark(Finder2(), 12, 12, reversed).report();
  FindRectBenchmark(Finder3(), 12, 12, reversed).report();
  FindRectBenchmark(Finder4(), 12, 12, reversed).report();

  print('\n');
  print('100');
  FindRectBenchmark(Finder3(), 12, 100, reversed).report();
  FindRectBenchmark(Finder4(), 12, 100, reversed).report();
  print('\n');
  print('100 - reversed');
  FindRectBenchmark(Finder3(), 12, 100, list).report();
  FindRectBenchmark(Finder4(), 12, 100, list).report();

  print('\n');
  print('1000');
  FindRectBenchmark(Finder3(), 12, 1000, list).report();
  FindRectBenchmark(Finder4(), 12, 1000, list).report();

  print('\n');
  print('1000 - reversed');
  FindRectBenchmark(Finder3(), 12, 1000, reversed).report();
  FindRectBenchmark(Finder4(), 12, 1000, reversed).report();

  print('\n');
  print('10000');
  FindRectBenchmark(Finder1(), 12, 10000, list).report();
  FindRectBenchmark(Finder2(), 12, 10000, list).report();
  FindRectBenchmark(Finder3(), 12, 10000, list).report();
  FindRectBenchmark(Finder4(), 12, 10000, list).report();

  print('\n');
  print('10000 - reversed');
  FindRectBenchmark(Finder1(), 12, 10000, reversed).report();
  FindRectBenchmark(Finder2(), 12, 10000, reversed).report();
  FindRectBenchmark(Finder3(), 12, 10000, reversed).report();
  FindRectBenchmark(Finder4(), 12, 10000, reversed).report();
}
