import 'dart:typed_data';

import 'package:example/unsplash_provider.dart';
import 'package:flutter/cupertino.dart';
import 'package:flutter/material.dart';
import 'package:flutter_blurhash/flutter_blurhash.dart';
import 'package:flutter_sticky_header/flutter_sticky_header.dart';
import 'package:intl/intl.dart';
import 'package:layout_grid/layout_grid.dart';

class UnsplashPhotoBuilder extends StatefulWidget {
  final Widget Function(List<UnsplashImage> images, Function(int) request)
      builder;

  const UnsplashPhotoBuilder({Key key, this.builder}) : super(key: key);

  @override
  _UnsplashPhotoBuilderState createState() => _UnsplashPhotoBuilderState();
}

class _UnsplashPhotoBuilderState extends State<UnsplashPhotoBuilder> {
  Future initialBuild;
  int lastPage = 1;
  final int perPage = 30; // 30 max
  List<UnsplashImage> images = [];

  @override
  void initState() {
    fetchImages(1);
    super.initState();
  }

  @override
  Widget build(BuildContext context) {
    return widget.builder(images, requestForImageInIndex);
  }

  void requestForImageInIndex(int index) {
    if (lastPage * perPage - 1 < index) {
      lastPage += 1;
      fetchImages(lastPage);
    }
  }

  Future<void> fetchImages(int page) async {
    lastPage = page;
    final List<UnsplashImage> result =
        await UnsplashImageProvider.loadImages(page: page, perPage: perPage);

    setState(() {
      images.addAll(result);
    });
  }
}

class CupertinoPhotos extends StatelessWidget {
  @override
  Widget build(BuildContext context) {
    return CupertinoTheme(
        data: CupertinoThemeData(brightness: Brightness.dark),
        child: CupertinoPageScaffold(
          child: Scaffold(
            backgroundColor: Colors.black,
            extendBody: true,
            bottomNavigationBar: CupertinoTabBar(
              items: [
                BottomNavigationBarItem(
                  icon: Icon(
                    Icons.photo_library,
                    size: 24,
                  ),
                  label: 'Library',
                ),
                BottomNavigationBarItem(
                  icon: Icon(
                    Icons.map,
                    size: 24,
                  ),
                  label: 'For you',
                ),
                BottomNavigationBarItem(
                  icon: Icon(
                    Icons.photo_album,
                    size: 24,
                  ),
                  label: 'Search',
                ),
                BottomNavigationBarItem(
                  icon: Icon(
                    Icons.search,
                    size: 24,
                  ),
                  label: 'Search',
                ),
              ],
            ),
            body: Scrollbar(
              child: UnsplashPhotoBuilder(
                builder: (list, requestIfNeeded) => CustomScrollView(
                  primary: true,
                  slivers: [
                    for (final int page in List.generate(20, (i) => i))
                      SliverStickyHeader(
                        overlapsContent: true,
                        header: Container(
                          height: 80.0,
                          decoration: BoxDecoration(
                              gradient: LinearGradient(
                                  begin: Alignment.topCenter,
                                  end: Alignment.bottomCenter,
                                  colors: [
                                Colors.black87,
                                Colors.black.withOpacity(0)
                              ],
                                  stops: [
                                0,
                                1
                              ])),
                          padding: EdgeInsets.symmetric(
                              horizontal: 20.0, vertical: 0),
                          alignment: Alignment.centerLeft,
                          child: Text(
                            DateFormat.yMMMd().format(
                              DateTime.now().add(Duration(days: -page)),
                            ),
                            style: Theme.of(context).accentTextTheme.headline5,
                          ),
                        ),
                        sliver: SliverLayoutGrid(
                          gridDelegate:
                              SliverLayoutGridDelegateWithFixedCrossAxisCount(
                            crossAxisCount: context.layout.value(
                              xs: 3,
                              md: 6,
                              lg: 9,
                            ),
                            crossAxisSpacing: 2,
                            mainAxisSpacing: 2,
                            masonry: true,
                          ),
                          delegate: SliverChildBuilderDelegate((c, j) {
                            final i = page * 16 + j;
                            requestIfNeeded(i + 16);
                            final image = list.length > i ? list[i] : null;
                            return Cell.aspectRatio(
                              key: ValueKey('Cell i + $page'),
                              snap: sizeForIndex(i),
                              child: Container(
                                child: image == null
                                    ? null
                                    : Stack(
                                        children: [
                                          if (image.blurHash != null)
                                            BlurHash(hash: image.blurHash),
                                          FadeInImage.memoryNetwork(
                                            fadeInDuration:
                                                Duration(milliseconds: 400),
                                            placeholder: kTransparentImage,
                                            image: image.getRegularUrl(),
                                            fit: BoxFit.cover,
                                            width: double.infinity,
                                            height: double.infinity,
                                          ),
                                        ],
                                      ),
                              ),
                            );
                          }, childCount: 16),
                        ),
                      ),
                  ],
                ),
              ),
            ),
          ),
        ));
  }

  int sizeForIndex(int index) {
    final int c = index % 16;
    if (c == 3) return 3;
    if (c == 6 || c == 14) return 2;
    return 1;
  }
}

//https://github.com/brianegan/transparent_image
final Uint8List kTransparentImage = new Uint8List.fromList(<int>[
  0x89,
  0x50,
  0x4E,
  0x47,
  0x0D,
  0x0A,
  0x1A,
  0x0A,
  0x00,
  0x00,
  0x00,
  0x0D,
  0x49,
  0x48,
  0x44,
  0x52,
  0x00,
  0x00,
  0x00,
  0x01,
  0x00,
  0x00,
  0x00,
  0x01,
  0x08,
  0x06,
  0x00,
  0x00,
  0x00,
  0x1F,
  0x15,
  0xC4,
  0x89,
  0x00,
  0x00,
  0x00,
  0x0A,
  0x49,
  0x44,
  0x41,
  0x54,
  0x78,
  0x9C,
  0x63,
  0x00,
  0x01,
  0x00,
  0x00,
  0x05,
  0x00,
  0x01,
  0x0D,
  0x0A,
  0x2D,
  0xB4,
  0x00,
  0x00,
  0x00,
  0x00,
  0x49,
  0x45,
  0x4E,
  0x44,
  0xAE,
]);
