import 'dart:convert';
import 'dart:io';

const String UNSPLASH_API_CLIENT_ID = String.fromEnvironment(
  'UNSPLASH_API_CLIENT_ID',
);

/// Helper class to interact with the Unsplash Api and provide [UnsplashImage].
class UnsplashImageProvider {
  /// Asynchronously loads a [UnsplashImage] for a given [id].
  static Future<UnsplashImage> loadImage(String id) async {
    String url = 'https://api.unsplash.com/photos/$id';
    // receive image data from unsplash
    var data = await _getImageData(url);
    // return image
    return UnsplashImage(data);
  }

  /// Asynchronously load a list of trending [UnsplashImage].
  /// Returns a list of [UnsplashImage].
  /// [page] is the page index for the api request.
  /// [perPage] sets the length of the returned list.
  static Future<List<UnsplashImage>> loadImages({int page = 1, int perPage = 10}) async {
    String url = 'https://api.unsplash.com/photos?page=$page&per_page=$perPage';
    // receive image data from unsplash
    var data = await _getImageData(url);
    // generate UnsplashImage List from received data
    List<UnsplashImage> images =
        List<UnsplashImage>.generate(data.length, (index) {
      return UnsplashImage(data[index]);
    });
    // return images
    return images;
  }

  /// Asynchronously load a list of [UnsplashImage] associated to a given [keyword].
  /// Returns a list where the first element is the [totalPages] available and the second element is a list of [UnsplashImage].
  /// [page] is the page index for the api request.
  /// [perPage] sets the length of the returned list.
  static Future<List> loadImagesWithKeyword(String keyword,
      {int page = 1, int perPage = 10}) async {
    // Search for image associated with the keyword
    String url =
        'https://api.unsplash.com/search/photos?query=$keyword&page=$page&per_page=$perPage&order_by=popular';
    // receive image data from unsplash associated to the given keyword
    var data = await _getImageData(url);
    // generate UnsplashImage List from received data
    List<UnsplashImage> images =
        List<UnsplashImage>.generate(data['results'].length, (index) {
      return UnsplashImage(data['results'][index]);
    });
    int totalPages = data['total_pages'];
    // return Images
    return [totalPages, images];
  }

  /// Receive image data from a given [url] and return the JSON decoded the data.
  static dynamic _getImageData(String url) async {
    // setup http client
    HttpClient httpClient = HttpClient();
    // setup http request
    HttpClientRequest request = await httpClient.getUrl(Uri.parse(url));
    // pass the client_id in the header
    request.headers.add('Authorization', 'Client-ID $UNSPLASH_API_CLIENT_ID');

    // wait for response
    HttpClientResponse response = await request.close();
    // Process the response
    if (response.statusCode == 200) {
      // response: OK
      // decode JSON
      String json = await response.transform(utf8.decoder).join();
      // return decoded json
      return jsonDecode(json);
    } else {
      // something went wrong :(
      print("Http error: ${response.statusCode}");
      // return empty list
      return [];
    }
  }
}

/// Model for Unsplash Image
class UnsplashImage {
  var data;

  UnsplashImage(this.data);

  // Getter

  String getId() {
    return data['id'];
  }

  String createdAt() {
    return data['created_at'];
  }

  String updatedAt() {
    return data['updated_at'];
  }

  int getWidth() {
    return data['width'];
  }

  int getHeight() {
    return data['height'];
  }

  String getColor() {
    return data['color'];
  }

  int getLikes() {
    return data['likes'];
  }

  bool isLikedByUser() {
    return data['liked_by_user'];
  }

  String getDescription() {
    return data['description'];
  }

  String getAltDescription() {
    return data['alt_description'];
  }

  getUrls() {
    return data['urls'];
  }

  String getRawUrl() {
    return getUrls()['raw'];
  }

  String getFullUrl() {
    return getUrls()['full'];
  }

  String getRegularUrl() {
    return getUrls()['regular'];
  }

  String getSmallUrl() {
    return getUrls()['small'];
  }

  String getThumbUrl() {
    return getUrls()['thumb'];
  }

  getLinks() {
    return data['links'];
  }

  String getSelfLink() {
    return getLinks()['self'];
  }

  String getHtmlLink() {
    return getLinks()['html'];
  }

  String getDownloadLink() {
    return getLinks()['download'];
  }

  String get blurHash {
    return data['blur_hash'];
  }
}
