import 'dart:math';

import 'package:flutter/physics.dart';
import 'package:flutter/widgets.dart';
import 'package:layout_grid/layout_grid.dart';

const double _epsilon = 0.0001;

/// Specifies how a grid is configured.
@immutable
class LayoutGridConfiguration {
  ///  Creates an object that holds the configuration of a layout grid.
  LayoutGridConfiguration({
    required this.rails,
    required this.railExtent,
    required this.mainAxisSpacing,
    required this.crossAxisSpacing,
    required this.reverseCrossAxis,
    required this.masonry,
    this.mainAxisOffsetsCacheSize = 2,
  })  : assert(rails > 0),
        assert(railExtent >= 0),
        assert(mainAxisSpacing >= 0),
        assert(crossAxisSpacing >= 0),
        assert(mainAxisOffsetsCacheSize > 0),
        crossAxisStride = railExtent + crossAxisSpacing;

  /// The maximum number of children in the cross axis.
  /// If the main axis is the vertical the rails would be the columns
  /// If the main axis is the horizontal the rails would be the rows
  final int rails;

  final bool masonry;

  /// The number of pixels from the leading edge of one cell to the trailing
  /// edge of the same cell in both axis.
  final double railExtent;

  /// The number of logical pixels between each child along the main axis.
  final double mainAxisSpacing;

  /// The number of logical pixels between each child along the cross axis.
  final double crossAxisSpacing;

  /// Whether the children should be placed in the opposite order of increasing
  /// coordinates in the cross axis.
  ///
  /// For example, if the cross axis is horizontal, the children are placed from
  /// left to right when [reverseCrossAxis] is false and from right to left when
  /// [reverseCrossAxis] is true.
  ///
  /// Typically set to the return value of [axisDirectionIsReversed] applied to
  /// the [SliverConstraints.crossAxisDirection].
  final bool reverseCrossAxis;

  final double crossAxisStride;

  /// The number of pages necessary to cache a mainAxisOffsets value.
  final int mainAxisOffsetsCacheSize;

  /// Computes the main axis extent of any layout cell.
  double? mainAxisExtentFor(CellData cell) {
    assert(cell.mainAxisExtent != null ||
        cell.mainAxisRelativeExtent != null ||
        cell.fitContent);

    if (cell.mainAxisExtent != null) {
      return cell.mainAxisExtent!;
    } else if (cell.fitContent) {
      return null;
    } else {
      final mainAxisRelativeExtent = cell.mainAxisRelativeExtent!;
      final spacing = mainAxisSpacing * (mainAxisRelativeExtent - 1);
      return mainAxisRelativeExtent * railExtent + spacing;
    }
  }

  double crossAxisOffsetFor(CellData data, _CellOffset offset) {
    final cellRailCount = data.railCount.clamp(0, rails);
    int railStartIndex = offset.startIndex + data.railLeadingOffset;
    if (railStartIndex + cellRailCount > rails) {
      railStartIndex = rails - cellRailCount;
    }
    if (reverseCrossAxis) {
      railStartIndex = rails - cellRailCount - railStartIndex;
    }
    return railStartIndex * crossAxisStride;
  }

  double crossAxisExtentFor(CellData data) {
    final cellRailCount = data.railCount.clamp(0, rails);
    return crossAxisStride * cellRailCount - crossAxisSpacing;
  }

  SliverLayoutGeometry childGeometry(
    CellParentData parentData,
    List<double> scrollOffsetByRail,
  ) {
    final cellData = parentData.data;
    final cellRailCount = cellData.railCount.clamp(0, rails);
    final cellRailStride = cellData.railStride.clamp(0, rails);
    final railLeadingOffset = cellRailStride - cellRailCount;
    assert(railLeadingOffset >= 0);

    final scrollOffsets = scrollOffsetByRail.map(
      (offset) => offset == 0 ? offset : offset + mainAxisSpacing,
    );

    final rect = _findFirstRectFor(cellRailStride, scrollOffsets);

    final double scrollOffset = rect.scrollOffset;
    final double? mainAxisExtent = mainAxisExtentFor(cellData);
    final double crossAxisOffset = crossAxisOffsetFor(cellData, rect);
    final double crossAxisExtent = crossAxisExtentFor(cellData);

    return SliverLayoutGeometry(
      scrollOffset: scrollOffset,
      crossAxisOffset: crossAxisOffset,
      mainAxisExtent: mainAxisExtent,
      crossAxisExtent: crossAxisExtent,
      railCount: cellRailCount,
      startRailIndex: rect.startIndex,
      railLeadingOffset: railLeadingOffset,
    );
  }

  /// Finds the first available rect with at least the specified [requiredRails].
  _CellOffset _findFirstRectFor(
      int requiredRails, Iterable<double> scrollOffsets,
      {int initialIndex = 0}) {
    assert(scrollOffsets.length != 0,
        'There should be at least one rail available');
    assert(scrollOffsets.length >= requiredRails,
        'Available rails are less than the ones requested');
    final List<double> offsets = scrollOffsets.toList();
    _CellOffset? _candidatePosition;
    _CellOffset? _possiblePosition;

    int startRailIndex = 0;
    int railCount = 1;
    bool contiguous = false;
    double minScrollOffset = double.infinity;
    double maxScrollOffset = 0;

    void setCandidateOffsetIfMeetsRequiremets() {
      if (_candidatePosition == null ||
          _candidatePosition!.scrollOffset > minScrollOffset) {
        _candidatePosition = _CellOffset(startRailIndex, minScrollOffset);
      }
    }

    void setPossibleOffsetIfMeetsRequirements() {
      if (_possiblePosition == null ||
          _possiblePosition!.scrollOffset > minScrollOffset) {
        _possiblePosition = _CellOffset(startRailIndex, minScrollOffset);
      }
    }

    // We have to use the nearEqual function because of floating-point arithmetic.
    // Ex: 0.1 + 0.2 = 0.30000000000000004 and not 0.3.
    for (var i = initialIndex; i < scrollOffsets.length; ++i) {
      final double offset = offsets[i];
      final bool equalOffset = nearEqual(offset, minScrollOffset, _epsilon);
      if (maxScrollOffset < offset) maxScrollOffset = offset;

      // This algorithm checks first if the current available rail is contiguous
      // to the previous one and they have the same scrollOffset. In that case
      // the rail count of the available space it is increased by one.
      //
      // In the contrary case, first it check that the previous space is valid
      // to be the candidate.
      //
      // Secondly if the current new rail is positioned with a minor offset than
      // the candidate we set the previous space as possible candidate
      // and we analyze the new space.
      // Otherwise if the new rail is positioned at a mayor offset we set the
      // contagious to false
      if (equalOffset && contiguous) {
        railCount++;
      } else {
        if (contiguous && railCount >= requiredRails) {
          setCandidateOffsetIfMeetsRequiremets();
        }
        if (!equalOffset && offset < minScrollOffset) {
          if (i != initialIndex && _candidatePosition == null) {
            setPossibleOffsetIfMeetsRequirements();
          }
          contiguous = true;
          startRailIndex = i;
          railCount = 1;
          minScrollOffset = offset;
        } else {
          contiguous = false;
        }
      }
    }

    if (railCount >= requiredRails) {
      return _CellOffset(startRailIndex, minScrollOffset);
    }

    if (_candidatePosition == null) {
      for (var i = 0; i < offsets.length; ++i) {
        offsets[i] =
            max(offsets[i], _possiblePosition?.scrollOffset ?? maxScrollOffset);
      }
      return _findFirstRectFor(
        requiredRails,
        offsets,
        // initialIndex: _possibleRect?.startRailIndex ?? 0,
      );
    } else {
      return _candidatePosition!;
    }
  }
}

class _CellOffset {
  const _CellOffset(
    this.startIndex,
    this.scrollOffset,
  );
  // Main axis
  final double scrollOffset;
  // Cross axis
  final int startIndex;
}
