import 'dart:collection';
import 'dart:math' as math;

import 'package:layout_grid/src/rendering/sliver_geometry.dart';
import 'package:layout_grid/src/rendering/sliver_section_cache_adaptor.dart';

import 'sliver_variable_size_box_adaptor.dart';
import '../widgets/cell.dart';
import 'package:flutter/foundation.dart';
import 'package:flutter/rendering.dart';

/// Describes the placement of a child in a [RenderSliverLayoutGrid].
///
/// See also:
///
///  * [RenderSliverLayoutGrid], which uses this class during its
///    [RenderSliverLayoutGrid.performLayout] method.
@immutable
class SliverLayoutGeometry {
  /// Creates an object that describes the placement of a child in a [RenderSliverGrid].
  const SliverLayoutGeometry({
    required this.scrollOffset,
    required this.crossAxisOffset,
    required this.mainAxisExtent,
    required this.crossAxisExtent,
    required this.railCount,
    required this.railLeadingOffset,
    required this.startRailIndex,
  });

  /// The scroll offset of the leading edge of the child relative to the leading
  /// edge of the parent.
  final double scrollOffset;

  /// The offset of the child in the non-scrolling axis.
  ///
  /// If the scroll axis is vertical, this offset is from the left-most edge of
  /// the parent to the left-most edge of the child. If the scroll axis is
  /// horizontal, this offset is from the top-most edge of the parent to the
  /// top-most edge of the child.
  final double crossAxisOffset;

  /// The extent of the child in the scrolling axis.
  ///
  /// If the scroll axis is vertical, this extent is the child's height. If the
  /// scroll axis is horizontal, this extent is the child's width.
  final double? mainAxisExtent;

  /// The extent of the child in the non-scrolling axis.
  ///
  /// If the scroll axis is vertical, this extent is the child's width. If the
  /// scroll axis is horizontal, this extent is the child's height.
  final double crossAxisExtent;

  /// The number of cells occupied in the cross axis.
  final int railCount;

  /// The number of rails between the first available position and the cell.
  final int railLeadingOffset;

  int get railStride => railCount + railLeadingOffset;

  final int startRailIndex;

  int get endRailIndex => startRailIndex + railCount - 1;

  /// The scroll offset of the trailing edge of the child relative to the
  /// leading edge of the parent.
  double? get trailingScrollOffset =>
      mainAxisExtent == null ? null : scrollOffset + mainAxisExtent!;

  /// Returns a tight [BoxConstraints] that forces the child to have the
  /// required size.
  BoxConstraints getBoxConstraints(SliverConstraints constraints) {
    return constraints.asBoxConstraints(
      minExtent: mainAxisExtent ?? 0,
      maxExtent: mainAxisExtent ?? double.infinity,
      crossAxisExtent: crossAxisExtent,
    );
  }

  @override
  String toString() {
    final List<String> properties = <String>[
      'scrollOffset: $scrollOffset',
      'crossAxisOffset: $crossAxisOffset',
      'mainAxisExtent: $mainAxisExtent',
      'crossAxisExtent: $crossAxisExtent',
      'crossAxisCellCount: $railCount',
      'startIndex: $startRailIndex'
    ];
    return 'SliverGridGeometry(${properties.join(', ')})';
  }

  SliverLayoutGeometry copyWith({
    double? scrollOffset,
    double? crossAxisOffset,
    double? mainAxisExtent,
    double? crossAxisExtent,
    int? crossAxisCellCount,
    int? startRailIndex,
    int? railLeadingOffset,
  }) {
    return SliverLayoutGeometry(
        scrollOffset: scrollOffset ?? this.scrollOffset,
        crossAxisOffset: crossAxisOffset ?? this.crossAxisOffset,
        mainAxisExtent: mainAxisExtent ?? this.mainAxisExtent,
        crossAxisExtent: crossAxisExtent ?? this.crossAxisExtent,
        railCount: crossAxisCellCount ?? this.railCount,
        startRailIndex: startRailIndex ?? this.startRailIndex,
        railLeadingOffset: railLeadingOffset ?? this.railLeadingOffset);
  }
}

class RenderCellData {
  final RenderBox box;
  final SliverLayoutGeometry geometry;
  final CellParentData parentData;

  RenderCellData({
    required this.box,
    required this.geometry,
  }) : parentData = box.parentData! as CellParentData;
}

/// A sliver that places multiple box children in a two dimensional arrangement.
///
/// [RenderSliverGrid] places its children in arbitrary positions determined by
/// [gridDelegate]. Each child is forced to have the size specified by the
/// [gridDelegate].
///
/// See also:
///
///  * [RenderSliverList], which places its children in a linear
///    array.
///  * [RenderSliverFixedExtentList], which places its children in a linear
///    array with a fixed extent in the main axis.
class RenderSliverLayoutGrid extends RenderSliverVariableSizeBoxAdaptor
    with PageCacheManagerMixin {
  /// Creates a sliver that contains multiple box children that whose size and
  /// position are determined by a delegate.
  ///
  /// The [configuration] and [childManager] arguments must not be null.
  RenderSliverLayoutGrid({
    required RenderSliverVariableSizeBoxChildManager childManager,
    required SliverLayoutGridDelegate gridDelegate,
  })  : _gridDelegate = gridDelegate,
        super(childManager: childManager);

  @override
  void setupParentData(RenderObject child) {
    if (child.parentData is! CellParentData) {
      child.parentData = CellParentData();
    }
  }

  /// A delegate that controls the configuration of the layout grid.
  SliverLayoutGridDelegate get gridDelegate => _gridDelegate;
  SliverLayoutGridDelegate _gridDelegate;
  set gridDelegate(SliverLayoutGridDelegate value) {
    if (_gridDelegate == value) return;
    if (value.runtimeType != _gridDelegate.runtimeType ||
        value.shouldRelayout(_gridDelegate)) markNeedsLayout();
    _gridDelegate = value;
  }

  /// Layouts the children that depend of other children in the same crossAxisOffset
  /// to calculate its mainAxisExtent.
  ///
  /// It calculates the maximun offset between all rails and layouts these
  /// children to fill the available space
  void performLayoutForChildrenWithIntrinsicExtent(
    List<RenderCellData> children,
    LayoutGridConfiguration configuration,
    List<double> railScrollOffsets,
  ) {
    assert(children.length <= railScrollOffsets.length,
        'There should not be more children than rails available');
    final trailingScrollOffset = railScrollOffsets.reduce(math.max);

    for (final child in children) {
      SliverLayoutGeometry geometry = child.geometry;
      final correction = trailingScrollOffset - geometry.trailingScrollOffset!;

      if (correction > 0) {
        geometry = geometry.copyWith(
          mainAxisExtent: geometry.mainAxisExtent! + correction,
        );
        final totalRailCount = math.min(
          railScrollOffsets.length,
          geometry.railStride,
        );
        for (var i = 0; i < totalRailCount; i++) {
          railScrollOffsets[i + geometry.startRailIndex] =
              geometry.trailingScrollOffset!;
        }
      }

      addAndLayoutChild(
          child.parentData.index!, geometry.getBoxConstraints(this.constraints),
          parentUsesSize: true);
    }
    children.clear();
  }

  // Sets all rails to have the same offset as the current maximun.
  // It is used when there is a forced breakline with `LayoutBreakLine`
  // or if it is not masonery mode whenever there is a breakline.
  void performBreakLine() {
    final trailingScrollOffset = scrollOffsets.reduce(math.max);
    for (var i = 0; i < scrollOffsets.length; i++) {
      scrollOffsets[i] = trailingScrollOffset;
    }
  }

  @override
  void performLayout() {
    childManager.didStartLayout();
    childManager.setDidUnderflow(false);

    final double scrollOffset =
        constraints.scrollOffset + constraints.cacheOrigin;
    assert(scrollOffset >= 0.0);
    final double remainingExtent = constraints.remainingCacheExtent;
    assert(remainingExtent >= 0.0);
    final double targetEndScrollOffset = scrollOffset + remainingExtent;

    bool reachedEnd = false;
    double trailingScrollOffset = 0.0;
    double leadingScrollOffset = double.infinity;
    bool visible = false;
    int firstIndex = 0;
    int lastIndex = 0;

    LayoutGridConfiguration configuration =
        _gridDelegate.getConfiguration(constraints);

    // Cells with intrinsicExtent set to true, needs to correct its layout
    // after the next cells are laid out. They will be saved in this list
    // and corrected by calling `performLayoutForChildrenWithItrinsicExtent`
    // after all cells in the same line are layouted.
    // Notice this correction will not affect other cell layouts as it will
    // only fill an empty trailing space
    List<RenderCellData> pendingLayoutChildren = <RenderCellData>[];

    startPageCache(configuration.rails, configuration.mainAxisOffsetsCacheSize);

    HashSet<int> visibleIndices = HashSet<int>();

    // Iterate through all children while they can be visible.
    for (var index = firstChildIndexForPage;
        scrollOffsets.any((o) => o <= targetEndScrollOffset);
        index++) {
      // Remember to layout this child
      final RenderBox? child = obtainChild(index);

      if (child == null) {
        // There are either no children, or we are past the end of all our children.
        reachedEnd = true;
        break;
      }

      // ParentData and geometry for this child
      final CellParentData childParentData = child.parentData as CellParentData;

      if (isBreakline(child)) {
        // If there is a forced break line, rail offsets are extended to the current
        // maxium offset. Previous child that requires
        performBreakLine();
        if (visible) {
          // In every break line, if they are visible, every child dependent of a
          // inherited extent is layouted
          performLayoutForChildrenWithIntrinsicExtent(
            pendingLayoutChildren,
            configuration,
            scrollOffsets,
          );
        }
        // TODO: Is it needed to layout LayoutBreakLine child?
        addAndLayoutChild(index, BoxConstraints.tight(Size.zero));
      } else {
        SliverLayoutGeometry childGeometry = configuration.childGeometry(
          childParentData,
          scrollOffsets,
        );
        childParentData.layoutOffset = childGeometry.scrollOffset;
        childParentData.crossAxisOffset = childGeometry.crossAxisOffset;
        assert(childParentData.index == index);

        // For children that don't have a mainAxisExtent (fit size), they are laid out
        // to calculate it.
        bool needsIntrinsicExtent = childGeometry.trailingScrollOffset == null;
        if (needsIntrinsicExtent) {
          // Layout the child to compute its trailingScrollOffset.
          final BoxConstraints constraints = tightConstraintsFor(
            crossAxisExtent: childGeometry.crossAxisExtent,
          );
          final double mainAxisExtent = intrinsicExtentOf(child, constraints)!;
          childGeometry =
              childGeometry.copyWith(mainAxisExtent: mainAxisExtent);
        }
        assert(childGeometry.mainAxisExtent != null);

        // Find first visible child from the zero-index
        if (!visible &&
            targetEndScrollOffset >= childGeometry.scrollOffset &&
            scrollOffset <= childGeometry.trailingScrollOffset!) {
          visible = true;
          leadingScrollOffset = childGeometry.scrollOffset;
          firstIndex = index;
        }

        // Layout child if is inside the visible zone
        if (visible) {
          // If it is an cell with an intrinsic extent, we postpone the
          // layout until all the cells in the same crossAxis are layouted.
          if (!childParentData.data.intrinsic) {
            addAndLayoutChild(
              index,
              childGeometry.getBoxConstraints(constraints),
              parentUsesSize: needsIntrinsicExtent,
            );
          } else {
            pendingLayoutChildren.add(
              RenderCellData(box: child, geometry: childGeometry),
            );
          }
        }

        if (visible) {
          if (indices.contains(index)) {
            visibleIndices.add(index);
          }
        }

        // Check if child exceed the limit of current page/viewport
        // and change to next one if that is the case
        if (childGeometry.trailingScrollOffset! >= page.trailingScrollOffset) {
          createNewPage(scrollOffsets, index);
        }

        assert(childGeometry.railStride <= scrollOffsets.length);
        for (int i = 0; i < childGeometry.railStride; i++) {
          scrollOffsets[i + childGeometry.startRailIndex] =
              childGeometry.trailingScrollOffset!;
        }
        trailingScrollOffset = scrollOffsets.reduce(math.max);
        lastIndex = index;

        // Check if widget is at the end of line and will cause a line break.
        final lineBreak = childGeometry.endRailIndex + 1 == configuration.rails;
        if (visible && lineBreak) {
          // In every break line, if they are visible, every child dependent of a
          // inherited extent is layouted
          performLayoutForChildrenWithIntrinsicExtent(
            pendingLayoutChildren,
            configuration,
            scrollOffsets,
          );
        }

        // If it is not masonry mode, all rail offsets are extended to the current
        // maximun offset, so a uniformed grid is formed.
        if (lineBreak && !configuration.masonry) {
          performBreakLine();
        }
      }
    }

    // In case last children with inherited extent didn't have a line break
    // we perform it again
    performLayoutForChildrenWithIntrinsicExtent(
      pendingLayoutChildren,
      configuration,
      scrollOffsets,
    );

    collectGarbage(visibleIndices);

    // Issue between slivers, needs scroll correction
    if (!visible) {
      if (scrollOffset > trailingScrollOffset) {
        leadingScrollOffset = trailingScrollOffset;
      } else if (leadingScrollOffset == double.infinity) {
        leadingScrollOffset = scrollOffset;
      } else {
        geometry = SliverGeometry.zero;
        childManager.didFinishLayout();
        return;
      }
    }

    // Calculate estimated max scroll offset and update geometry
    double estimatedMaxScrollOffset;
    if (reachedEnd) {
      estimatedMaxScrollOffset = trailingScrollOffset;
    } else {
      estimatedMaxScrollOffset = childManager.estimateMaxScrollOffset(
        constraints,
        firstIndex: firstIndex,
        lastIndex: lastIndex,
        leadingScrollOffset: leadingScrollOffset,
        trailingScrollOffset: trailingScrollOffset,
      );
      assert(
        estimatedMaxScrollOffset >= trailingScrollOffset - leadingScrollOffset,
      );
    }
    assert(leadingScrollOffset <= trailingScrollOffset);

    final double paintExtent = calculatePaintOffset(
      constraints,
      from: leadingScrollOffset,
      to: trailingScrollOffset,
    );
    final double cacheExtent = calculateCacheOffset(
      constraints,
      from: leadingScrollOffset,
      to: trailingScrollOffset,
    );

    geometry = SliverGeometry(
      scrollExtent: estimatedMaxScrollOffset,
      paintExtent: paintExtent,
      cacheExtent: cacheExtent,
      maxPaintExtent: estimatedMaxScrollOffset,
      // Conservative to avoid flickering away the clip during scroll.
      hasVisualOverflow: trailingScrollOffset > targetEndScrollOffset ||
          constraints.scrollOffset > 0.0,
    );

    // We may have started the layout while scrolled to the end, which would not
    // expose a new child.
    if (estimatedMaxScrollOffset == trailingScrollOffset)
      childManager.setDidUnderflow(true);
    childManager.didFinishLayout();
  }

  bool isBreakline(RenderBox child) {
    final CellParentData childParentData = child.parentData as CellParentData;
    return childParentData.data.breakLine;
  }

  bool get debugPaintRailsEnabled => debugPaintSizeEnabled;
  bool get debugPaintGuttersEnabled => debugPaintBaselinesEnabled;

  @override
  void debugPaint(PaintingContext context, Offset offset) {
    super.debugPaint(context, offset);
    assert(() {
      final Canvas canvas = context.canvas;

      LayoutGridConfiguration configuration =
          _gridDelegate.getConfiguration(constraints);

      void paintRail(double crossAxisStart, double crossAxisEnd, Paint paint) {
        switch (constraints.axis) {
          case Axis.vertical:
            canvas.drawRect(
              Rect.fromPoints(
                offset.translate(crossAxisStart, 0),
                offset.translate(crossAxisEnd, geometry!.paintExtent),
              ),
              paint,
            );
            break;
          case Axis.horizontal:
            canvas.drawLine(
              offset.translate(0.0, crossAxisStart),
              offset.translate(geometry!.paintExtent, crossAxisEnd),
              paint,
            );
            break;
        }
      }

      if (debugPaintRailsEnabled) {
        final Paint paint = Paint()
          ..color = Color(0xFFFF00000).withOpacity(0.2)
          ..style = PaintingStyle.fill;
        for (var index = 0; index < configuration.rails; index++) {
          final crossAxisStart = index * configuration.crossAxisStride;
          paintRail(
              crossAxisStart, crossAxisStart + configuration.railExtent, paint);
        }
      }
      if (debugPaintGuttersEnabled) {
        final Paint paint = Paint()
          ..color = Color(0xFF0000FF).withOpacity(0.2)
          ..style = PaintingStyle.fill;
        for (var index = 0; index < configuration.rails - 1; index++) {
          final crossAxisStart =
              configuration.railExtent + index * configuration.crossAxisStride;
          paintRail(crossAxisStart,
              crossAxisStart + configuration.mainAxisSpacing, paint);
        }
      }
      return true;
    }());
  }
}

/// Creates a grid layout.
///
/// This delegate creates grids with variable sized but equally spaced cells.
///
/// See also:
///
///  * [LayoutGrid], which can use this delegate to control the layout of its
///    cells.
///  * [SliverLayoutGrid], which can use this delegate to control the layout of its
///    cells.
///  * [RenderSliverLayoutGrid], which can use this delegate to control the layout of
///    its cells.
abstract class SliverLayoutGridDelegate {
  /// Creates a delegate that makes grid layouts
  ///
  /// All of the arguments must not be null. The [mainAxisSpacing] and
  /// [crossAxisSpacing] arguments must not be negative.
  const SliverLayoutGridDelegate({
    this.mainAxisSpacing: 0.0,
    this.crossAxisSpacing: 0.0,
  })  : assert(mainAxisSpacing >= 0),
        assert(crossAxisSpacing >= 0);

  /// The number of logical pixels between each child along the main axis.
  final double mainAxisSpacing;

  /// The number of logical pixels between each child along the cross axis.
  final double crossAxisSpacing;

  bool _debugAssertIsValid() {
    assert(mainAxisSpacing >= 0.0);
    assert(crossAxisSpacing >= 0.0);
    return true;
  }

  /// Returns information about the layout grid configuration.
  LayoutGridConfiguration getConfiguration(SliverConstraints constraints);

  /// Override this method to return true when the children need to be
  /// laid out.
  ///
  /// This should compare the fields of the current delegate and the given
  /// `oldDelegate` and return true if the fields are such that the layout would
  /// be different.
  bool shouldRelayout(SliverLayoutGridDelegate oldDelegate) {
    return oldDelegate.mainAxisSpacing != mainAxisSpacing ||
        oldDelegate.crossAxisSpacing != crossAxisSpacing;
  }
}

/// Creates grid layouts with a fixed number of cells in the cross
/// axis.
///
/// For example, if the grid is vertical, this delegate will create a layout
/// with a fixed number of columns. If the grid is horizontal, this delegate
/// will create a layout with a fixed number of rows.
///
/// This delegate creates grids with variable sized but equally spaced cells.
///
/// See also:
///
///  * [SliverLayoutGridDelegate], which creates grid layouts.
///  * [LayoutGrid], which can use this delegate to control the layout of its
///    cells.
///  * [SliverLayoutGrid], which can use this delegate to control the layout of its
///    cells.
///  * [RenderSliverLayoutGrid], which can use this delegate to control the layout of
///    its cells.
class SliverLayoutGridDelegateWithFixedCrossAxisCount
    extends SliverLayoutGridDelegate {
  /// Creates a delegate that makes grid layouts with a fixed number
  /// of cells in the cross axis.
  ///
  /// All of the arguments must not be null. The [mainAxisSpacing] and
  /// [crossAxisSpacing] arguments must not be negative. The [crossAxisCount]
  /// argument must be greater than zero.
  const SliverLayoutGridDelegateWithFixedCrossAxisCount({
    required this.crossAxisCount,
    double mainAxisSpacing: 0.0,
    double crossAxisSpacing: 0.0,
    this.masonry = false,
  })  : assert(crossAxisCount > 0),
        super(
          mainAxisSpacing: mainAxisSpacing,
          crossAxisSpacing: crossAxisSpacing,
        );

  /// The number of children in the cross axis.
  final int crossAxisCount;

  final bool masonry;

  @override
  bool _debugAssertIsValid() {
    assert(crossAxisCount > 0);
    return super._debugAssertIsValid();
  }

  @override
  LayoutGridConfiguration getConfiguration(SliverConstraints constraints) {
    assert(_debugAssertIsValid());
    final double usableCrossAxisExtent =
        constraints.crossAxisExtent - crossAxisSpacing * (crossAxisCount - 1);
    final double railExtent = usableCrossAxisExtent / crossAxisCount;
    return LayoutGridConfiguration(
      rails: crossAxisCount,
      railExtent: railExtent,
      mainAxisSpacing: mainAxisSpacing,
      crossAxisSpacing: crossAxisSpacing,
      masonry: masonry,
      reverseCrossAxis: axisDirectionIsReversed(constraints.crossAxisDirection),
    );
  }

  @override
  bool shouldRelayout(
      covariant SliverLayoutGridDelegateWithFixedCrossAxisCount oldDelegate) {
    return oldDelegate.crossAxisCount != crossAxisCount ||
        super.shouldRelayout(oldDelegate);
  }
}

/// Create grid layouts with cells that each have a maximum
/// cross-axis extent.
///
/// This delegate will select a cross-axis extent for the cells that is as
/// large as possible subject to the following conditions:
///
///  - The extent evenly divides the cross-axis extent of the grid.
///  - The extent is at most [maxCrossAxisExtent].
///
/// For example, if the grid is vertical, the grid is 500.0 pixels wide, and
/// [maxCrossAxisExtent] is 150.0, this delegate will create a grid with 4
/// columns that are 125.0 pixels wide.
///
/// This delegate creates grids with variable sized but equally spaced cells.
///
/// See also:
///
///  * [SliverLayoutGridDelegate], which creates grid layouts.
///  * [LayoutGrid], which can use this delegate to control the layout of its
///    cells.
///  * [SliverLayoutGrid], which can use this delegate to control the layout of its
///    cells.
///  * [RenderSliverLayoutGrid], which can use this delegate to control the layout of
///    its cells.
class SliverLayoutGridDelegateWithMaxCrossAxisExtent
    extends SliverLayoutGridDelegate {
  /// Creates a delegate that makes grid layouts with cells that
  /// have a maximum cross-axis extent.
  ///
  /// All of the arguments must not be null. The [maxCrossAxisExtent],
  /// [mainAxisSpacing] and [crossAxisSpacing] arguments must not be negative.
  const SliverLayoutGridDelegateWithMaxCrossAxisExtent({
    required this.maxCrossAxisExtent,
    double mainAxisSpacing: 0.0,
    double crossAxisSpacing: 0.0,
    this.masonry = false,
  })  : assert(maxCrossAxisExtent > 0),
        super(
          mainAxisSpacing: mainAxisSpacing,
          crossAxisSpacing: crossAxisSpacing,
        );

  /// The maximum extent of cells in the cross axis.
  ///
  /// This delegate will select a cross-axis extent for the cells that is as
  /// large as possible subject to the following conditions:
  ///
  ///  - The extent evenly divides the cross-axis extent of the grid.
  ///  - The extent is at most [maxCrossAxisExtent].
  ///
  /// For example, if the grid is vertical, the grid is 500.0 pixels wide, and
  /// [maxCrossAxisExtent] is 150.0, this delegate will create a grid with 4
  /// columns that are 125.0 pixels wide.
  final double maxCrossAxisExtent;

  final bool masonry;

  @override
  bool _debugAssertIsValid() {
    assert(maxCrossAxisExtent >= 0);
    return super._debugAssertIsValid();
  }

  @override
  LayoutGridConfiguration getConfiguration(SliverConstraints constraints) {
    assert(_debugAssertIsValid());
    final int crossAxisCount =
        ((constraints.crossAxisExtent + crossAxisSpacing) /
                (maxCrossAxisExtent + crossAxisSpacing))
            .ceil();

    final double usableCrossAxisExtent =
        constraints.crossAxisExtent - crossAxisSpacing * (crossAxisCount - 1);

    final double cellExtent = usableCrossAxisExtent / crossAxisCount;
    return LayoutGridConfiguration(
      rails: crossAxisCount,
      masonry: masonry,
      railExtent: cellExtent,
      mainAxisSpacing: mainAxisSpacing,
      crossAxisSpacing: crossAxisSpacing,
      reverseCrossAxis: axisDirectionIsReversed(constraints.crossAxisDirection),
    );
  }

  @override
  bool shouldRelayout(
      covariant SliverLayoutGridDelegateWithMaxCrossAxisExtent oldDelegate) {
    return oldDelegate.maxCrossAxisExtent != maxCrossAxisExtent ||
        super.shouldRelayout(oldDelegate);
  }
}
