import 'dart:collection';
import 'package:flutter/rendering.dart';

typedef _PageCacheMap = SplayTreeMap<int, _PageCache>;

mixin PageCacheManagerMixin on RenderSliver {
  final HashMap<int, _PageCacheMap> _pagesCacheBySizeAndRails =
      HashMap<int, _PageCacheMap>();

  // If the viewport is resized, we keep the in memory the old offsets caches. (Useful if only the orientation changes multiple times).
  _PageCacheMap? _pages;
  _PageCacheMap get pages {
    assert(_pages != null, 'startPageCache(rails) not called');
    return _pages!;
  }

  _PageCache? _page;
  _PageCache get page {
    assert(_page != null, 'startPageCache(rails) not called');
    return _page!;
  }

  int get firstChildIndexForPage => page.firstChildIndex;

  double? _pageExtent;
  double get pageExtent {
    assert(_pageExtent != null, 'startPageCache(rails) not called');
    return _pageExtent!;
  }

  // Scroll Offsets by rails
  List<double>? _scrollOffsets;
  List<double> get scrollOffsets {
    assert(_scrollOffsets != null, 'startPageCache(rails) not called');
    return _scrollOffsets!;
  }

  void startPageCache(int rails, int cachePageCount) {
    final scrollOffset = constraints.scrollOffset + constraints.cacheOrigin;

    _pageExtent = cachePageCount * constraints.viewportMainAxisExtent;
    _pages = _pagesCacheBySizeAndRails.putIfAbsent(
      Object.hash(pageExtent, rails),
      () => _PageCacheMap(),
    );

    if (pages.isEmpty) {
      _page = pages.putIfAbsent(0, () => _PageCache.empty(rails, pageExtent));
    } else {
      int pageIndex = scrollOffset ~/ pageExtent;
      assert(pageIndex >= 0);
      final int smallestKey = pages.lastKeyBefore(pageIndex + 1)!;
      _page = pages[smallestKey]!;
    }

    _scrollOffsets = page.railOffsets.toList();
  }

  void createNewPage(List<double> scrollOffsets, int firstChildIndex) {
    int nextPageIndex = page.index + 1;
    final nextSectionOffset = _PageCache(
      scrollOffsets,
      (nextPageIndex + 1.0) * pageExtent,
      nextPageIndex,
      firstChildIndex,
    );
    pages[nextPageIndex] = nextSectionOffset;
    _page = nextSectionOffset;
  }
}

class _PageCache {
  _PageCache(
    List<double> railOffsets,
    this.trailingScrollOffset, [
    this.index = 0,
    this.firstChildIndex = 0,
  ]) : this.railOffsets = List.from(railOffsets);

  _PageCache.empty(
    int rails,
    this.trailingScrollOffset, [
    this.index = 0,
    this.firstChildIndex = 0,
  ]) : this.railOffsets = List.filled(rails, 0.0);

  final int index;

  final int firstChildIndex;

  final double trailingScrollOffset;

  final List<double> railOffsets;

  @override
  String toString() =>
      '[$index-$trailingScrollOffset] ($firstChildIndex, $railOffsets)';
}
