import 'sliver_layout_grid_view.dart';

import '../rendering/sliver_variable_size_box_adaptor.dart';
import '../rendering/sliver_layout_grid.dart';
import 'package:flutter/foundation.dart';
import 'package:flutter/material.dart';
import 'package:flutter/rendering.dart';

class Cell extends ParentDataWidget<CellParentData> {
  Cell({
    Key? key,
    required Widget child,
    int snap = 4,
    int offset = 0,
    required double extent,
  })  : data = CellData.extent(
          snap,
          extent,
          railLeadingOffset: offset,
        ),
        super(key: key, child: child);

  Cell.fit({
    Key? key,
    required Widget child,
    int snap = 4,
    int offset = 0,
  })  : data = CellData.fit(
          snap,
          railLeadingOffset: offset,
        ),
        super(key: key, child: child);

  Cell.relativeExtent({
    Key? key,
    int snap = 4,
    double? relativeExtent,
    required Widget child,
    int offset = 0,
  })  : data = CellData.count(
          snap,
          relativeExtent ?? snap.toDouble(),
          railLeadingOffset: offset,
        ),
        super(key: key, child: child);

  Cell.aspectRatio({
    Key? key,
    int snap = 4,
    double aspectRatio = 1,
    required Widget child,
    int offset = 0,
  })  : data = CellData.count(
          snap,
          aspectRatio * snap,
          railLeadingOffset: offset,
        ),
        super(key: key, child: child);

  const Cell.breakLine({
    Key? key,
  })  : data = const CellData.breakLine(),
        super(key: key, child: const SizedBox.shrink());

  Cell._({
    Key? key,
    required int snap,
    required Widget child,
    bool inheritMainAxisExtent = false,
    int offset = 0,
  })  : data = CellData.fit(
          snap,
          intrinsic: inheritMainAxisExtent,
          railLeadingOffset: offset,
        ),
        super(key: key, child: child);

  final CellData data;

  /// Returns the first render object that complies with the following:
  ///   - Its parent data is `CellParentData`
  ///   - Its parent is a `RenderSliverLayoutGrid`
  ///   - If is an renderObject with less depth that the `renderObject`
  ///   it can only have a  `RenderRepaintBoundary` or `RenderIndexedSemantics`
  ///   in between or a `RenderRepaintBoundary` and `RenderIndexedSemantics` in
  ///   that order. These render objects could be  added by
  ///   the SliverLayoutGrid for accesibility and performance
  ///  Notice this methid can return if it can't finde a render object that
  ///  complies with these conditions
  RenderObject? firstValidRenderObject(RenderObject renderObject) {
    RenderObject targetObject = renderObject;

    bool isValidRenderObject(RenderObject renderObject) {
      return renderObject.parent is RenderSliverLayoutGrid &&
          renderObject.parentData is CellParentData;
    }

    // First check, true if there is no `RenderRepaintBoundary` and `RenderIndexedSemantics`
    // and renderObject is valid
    if (isValidRenderObject(targetObject)) {
      return targetObject;
    } else {
      targetObject = renderObject.parent as RenderObject;
    }

    // Second check, true if there is a `RenderRepaintBoundary` and its
    // parent is valid
    if (targetObject is RenderRepaintBoundary &&
        isValidRenderObject(targetObject)) {
      return targetObject;
    } else if (targetObject is! RenderIndexedSemantics) {
      targetObject = targetObject.parent as RenderObject;
    }

    // Third check, true if there is a `RenderIndexedSemantics` and its
    // parent is valid
    // It can be nested [RenderRepaintBoundary/RenderIndexedSemantics/TargetRenderObject]
    if (targetObject is RenderIndexedSemantics &&
        isValidRenderObject(targetObject)) {
      return targetObject;
    }

    return null;
  }

  /// Between this widget and the target RenderObject there could be a
  /// `RenderRepaintBoundary` and `RenderIndexedSemantics` added by
  /// the SliverLayoutGrid for accesibility and performance
  bool debugIsValidRenderObject(RenderObject renderObject) {
    return firstValidRenderObject(renderObject) != null;
  }

  @override
  void applyParentData(RenderObject renderObject) {
    RenderObject? targetObjet = firstValidRenderObject(renderObject)!;
    final CellParentData parentData = targetObjet.parentData! as CellParentData;
    bool needsLayout = false;

    if (parentData.data != data) {
      parentData.data = data;
      needsLayout = true;
    }

    if (needsLayout) {
      final AbstractNode? targetParent = targetObjet.parent;
      if (targetParent is RenderObject) targetParent.markNeedsLayout();
    }
  }

  @override
  Type get debugTypicalAncestorWidgetClass => SliverLayoutGrid;

  @override
  void debugFillProperties(DiagnosticPropertiesBuilder properties) {
    super.debugFillProperties(properties);
    // properties.add(IntProperty('span', span));
  }
}

/// Holds the dimensions of a [Cell].
///
/// A [CellData] always overlaps an exact number of cells in the cross
/// axis of a [LayoutGrid].
/// The main axis extent can either be a number of pixels or a number of
/// cells to overlap.
class CellData {
  /// Creates a [CellData] with the given [railCount] and
  /// [mainAxisRelativeExtent].
  ///
  /// The main axis extent of this cell will be the length of
  /// [mainAxisRelativeExtent] cells (inner spacings included).
  const CellData.count(
    this.railCount,
    double mainAxisRelativeExtent, {
    this.railLeadingOffset = 0,
  })  : assert(railCount >= 0),
        assert(mainAxisRelativeExtent >= 0),
        this.mainAxisRelativeExtent = mainAxisRelativeExtent,
        mainAxisExtent = null,
        intrinsic = false,
        breakLine = false;

  /// Creates a [CellData] with the given [railCount] and
  /// [mainAxisExtent].
  ///
  /// This cell will have a fixed main axis extent.
  const CellData.extent(
    this.railCount,
    double mainAxisExtent, {
    this.railLeadingOffset = 0,
  })  : assert(railCount >= 0),
        assert(mainAxisExtent >= 0),
        this.mainAxisExtent = mainAxisExtent,
        mainAxisRelativeExtent = null,
        intrinsic = false,
        breakLine = false;

  /// Creates a [CellData] with the given [railCount] that
  /// fit its main axis extent to its content.
  ///
  /// This cell will have a fixed main axis extent.
  const CellData.fit(
    this.railCount, {
    this.intrinsic = false,
    this.railLeadingOffset = 0,
  })  : assert(railCount >= 0),
        mainAxisExtent = null,
        mainAxisRelativeExtent = null,
        breakLine = false;

  /// Creates a [CellData] with the given [railCount] that
  /// fit its main axis extent to its content.
  ///
  /// This cell will have a fixed main axis extent.
  const CellData.breakLine()
      : this.intrinsic = false,
        mainAxisExtent = null,
        railCount = 0,
        mainAxisRelativeExtent = null,
        railLeadingOffset = 0,
        breakLine = true;

  /// The number of pixels occupied in the main axis.
  final double? mainAxisExtent;

  /// The number of cells occupied in the main axis.
  final num? mainAxisRelativeExtent;

  final bool intrinsic;

  /// The number of cells occupied in the cross axis.
  final int railCount;

  /// The number of rails between the first available position and the cell.
  final int railLeadingOffset;

  int get railStride => railCount + railLeadingOffset;

  final bool breakLine;

  bool get fitContent =>
      mainAxisRelativeExtent == null && mainAxisExtent == null;
}

class CellParentData extends SliverVariableSizeBoxAdaptorParentData {
  CellData data = CellData.count(1, 1);
}

class LayoutBreakLine extends Cell {
  const LayoutBreakLine() : super.breakLine();
}

class IntrinsicCell extends Cell {
  IntrinsicCell({
    Key? key,
    int snap = 4,
    required Widget child,
    bool inheritExtent = true,
    int offset = 0,
  }) : super._(
          key: key,
          child: child,
          snap: snap,
          offset: offset,
          inheritMainAxisExtent: inheritExtent,
        );
}
