import 'package:layout/layout.dart';

import '../rendering/sliver_layout_grid.dart';
import 'sliver_layout_grid_view.dart';

import 'package:flutter/foundation.dart';
import 'package:flutter/widgets.dart';

/// A scrollable, 2D array of widgets with variable sizes.
///
/// The main axis direction of a grid is the direction in which it scrolls (the
/// [scrollDirection]).
///
/// The most commonly used grid layouts are [LayoutGrid.count], which
/// creates a layout with a fixed number of cells in the cross axis, and
/// [LayoutGrid.extent], which creates a layout with cells that have a maximum
/// cross-axis extent. A custom [SliverLayoutGridDelegate] can produce an
/// arbitrary 2D arrangement of children.
///
/// To create a grid with a large (or infinite) number of children, use the
/// [LayoutGrid.builder] constructor with either a
/// [SliverLayoutGridDelegateWithFixedCrossAxisCount] or a
/// [SliverLayoutGridDelegateWithMaxCrossAxisExtent] for the [gridDelegate].
/// You can also use the [LayoutGrid.countBuilder] or
/// [LayoutGrid.extentBuilder] constructors.
///
/// To use a custom [SliverVariableSizeChildDelegate], use [LayoutGrid.custom].
///
/// To create a linear array of children, use a [ListView].
///
/// To control the initial scroll offset of the scroll view, provide a
/// [controller] with its [ScrollController.initialScrollOffset] property set.
///
/// ### Sample code
///
/// Here are two brief snippets showing a [LayoutGrid] and its equivalent using
/// [CustomScrollView]:
///
/// ```dart
///LayoutGrid.count(
///   primary: false,
///   crossAxisCount: 4,
///   mainAxisSpacing: 4.0,
///   crossAxisSpacing: 4.0,
///   children: const <Widget>[
///     const Cell.fit(child: Text('1')),
///     const Cell.fit(child: Text('2')),
///     const Cell.fit(child: Text('3')),
///     const Cell.fit(child: Text('4')),
///     const Cell.fit(child: Text('5')),
///     const Cell.fit(child: Text('6')),
///     const Cell.fit(child: Text('7')),
///     const Cell.fit(child: Text('8')),
///   ],
/// )
/// ```
///
/// ```dart
///CustomScrollView(
///   primary: false,
///   slivers: <Widget>[
///    SliverLayoutGrid.count(
///       crossAxisCount: 4,
///       mainAxisSpacing: 4.0,
///       crossAxisSpacing: 4.0,
///       children: const <Widget>[
///         const Cell(child: Text('1')),
///         const Cell(child: Text('2')),
///         const Cell(child: Text('3')),
///         const Cell(child: Text('4')),
///         const Cell(child: Text('5')),
///         const Cell(child: Text('6')),
///         const Cell(child: Text('7')),
///         const Cell(child: Text('8')),
///       ],
///     )
///   ],
/// )
/// ```
///
/// See also:
///
///  * [SingleChildScrollView], which is a scrollable widget that has a single
///    child.
///  * [ListView], which is scrollable, linear list of widgets.
///  * [PageView], which is a scrolling list of child widgets that are each the
///    size of the viewport.
///  * [CustomScrollView], which is a scrollable widget that creates custom
///    scroll effects using slivers.
///  * [SliverLayoutGridDelegateWithFixedCrossAxisCount], which creates a
///    layout with a fixed number of cells in the cross axis.
///  * [SliverLayoutGridDelegateWithMaxCrossAxisExtent], which creates a
///    layout with cells that have a maximum cross-axis extent.
///  * [ScrollNotification] and [NotificationListener], which can be used to watch
///    the scroll position without using a [ScrollController].
class LayoutGrid extends BoxScrollView {
  /// Creates a scrollable, 2D array of widgets with a custom
  /// [SliverLayoutGridDelegate].
  ///
  /// The [gridDelegate] argument must not be null.
  ///
  /// The `addAutomaticKeepAlives` argument corresponds to the
  /// [SliverLayoutGridDelegate.addAutomaticKeepAlives] property. The
  /// `addRepaintBoundaries` argument corresponds to the
  /// [SliverLayoutGridDelegate.addRepaintBoundaries] property. Both must not be
  /// null.
  LayoutGrid({
    Key? key,
    Axis scrollDirection: Axis.vertical,
    bool reverse: false,
    ScrollController? controller,
    bool? primary,
    ScrollPhysics? physics,
    bool shrinkWrap: false,
    EdgeInsetsGeometry? padding,
    this.gridDelegate,
    bool addAutomaticKeepAlives: true,
    bool addRepaintBoundaries: true,
    List<Widget> children: const <Widget>[],
  })  : childrenDelegate = SliverChildListDelegate(
          children,
          addAutomaticKeepAlives: addAutomaticKeepAlives,
          addRepaintBoundaries: addRepaintBoundaries,
        ),
        super(
          key: key,
          scrollDirection: scrollDirection,
          reverse: reverse,
          controller: controller,
          primary: primary,
          physics: physics,
          shrinkWrap: shrinkWrap,
          padding: padding,
        );

  /// Creates a scrollable, 2D array of widgets that are created on demand.
  ///
  /// This constructor is appropriate for grid views with a large (or infinite)
  /// number of children because the builder is called only for those children
  /// that are actually visible.
  ///
  /// Providing a non-null [itemCount] improves the ability of the
  /// [SliverLayoutGridDelegate] to estimate the maximum scroll extent.
  ///
  /// [itemBuilder] will be called only with indices greater than or equal to
  /// zero and less than [itemCount].
  ///
  /// The [gridDelegate] argument must not be null.
  ///
  /// The `addAutomaticKeepAlives` argument corresponds to the
  /// [SliverVariableSizeChildBuilderDelegate.addAutomaticKeepAlives] property. The
  /// `addRepaintBoundaries` argument corresponds to the
  /// [SliverVariableSizeChildBuilderDelegate.addRepaintBoundaries] property. Both must not
  /// be null.
  LayoutGrid.builder({
    Key? key,
    Axis scrollDirection: Axis.vertical,
    bool reverse: false,
    ScrollController? controller,
    bool? primary,
    ScrollPhysics? physics,
    bool shrinkWrap: false,
    EdgeInsetsGeometry? padding,
    this.gridDelegate,
    required IndexedWidgetBuilder itemBuilder,
    int? itemCount,
    bool addAutomaticKeepAlives: true,
    bool addRepaintBoundaries: true,
  })  : childrenDelegate = SliverChildBuilderDelegate(
          itemBuilder,
          childCount: itemCount,
          addAutomaticKeepAlives: addAutomaticKeepAlives,
          addRepaintBoundaries: addRepaintBoundaries,
        ),
        super(
          key: key,
          scrollDirection: scrollDirection,
          reverse: reverse,
          controller: controller,
          primary: primary,
          physics: physics,
          shrinkWrap: shrinkWrap,
          padding: padding,
        );

  /// Creates a scrollable, 2D array of widgets with both a custom
  /// [SliverLayoutGridDelegate] and a custom [SliverVariableSizeChildDelegate].
  ///
  /// To use an [IndexedWidgetBuilder] callback to build children, either use
  /// a [SliverVariableSizeChildBuilderDelegate] or use the
  /// [SliverLayoutGridDelegate.builder] constructor.
  ///
  /// The [gridDelegate] and [childrenDelegate] arguments must not be null.
  LayoutGrid.custom({
    Key? key,
    Axis scrollDirection: Axis.vertical,
    bool reverse: false,
    ScrollController? controller,
    bool? primary,
    ScrollPhysics? physics,
    bool shrinkWrap: false,
    EdgeInsetsGeometry? padding,
    this.gridDelegate,
    required this.childrenDelegate,
  }) : super(
          key: key,
          scrollDirection: scrollDirection,
          reverse: reverse,
          controller: controller,
          primary: primary,
          physics: physics,
          shrinkWrap: shrinkWrap,
          padding: padding,
        );

  /// Creates a scrollable, 2D array of widgets of variable sizes with a fixed
  /// number of cells in the cross axis.
  ///
  /// Uses a [SliverLayoutGridDelegateWithFixedCrossAxisCount] as the
  /// [gridDelegate].
  ///
  /// The `addAutomaticKeepAlives` argument corresponds to the
  /// [SliverLayoutGridDelegate.addAutomaticKeepAlives] property. The
  /// `addRepaintBoundaries` argument corresponds to the
  /// [SliverLayoutGridDelegate.addRepaintBoundaries] property. Both must not be
  /// null.
  ///
  /// See also:
  ///
  ///  * [new SliverGrid.count], the equivalent constructor for [SliverGrid].
  LayoutGrid.count({
    Key? key,
    Axis scrollDirection: Axis.vertical,
    bool reverse: false,
    ScrollController? controller,
    bool? primary,
    ScrollPhysics? physics,
    bool shrinkWrap: false,
    EdgeInsetsGeometry? padding,
    required int crossAxisCount,
    double mainAxisSpacing: 0.0,
    double crossAxisSpacing: 0.0,
    bool addAutomaticKeepAlives: true,
    bool addRepaintBoundaries: true,
    List<Widget> children: const <Widget>[],
  })  : gridDelegate = SliverLayoutGridDelegateWithFixedCrossAxisCount(
          crossAxisCount: crossAxisCount,
          mainAxisSpacing: mainAxisSpacing,
          crossAxisSpacing: crossAxisSpacing,
        ),
        childrenDelegate = SliverChildListDelegate(
          children,
          addAutomaticKeepAlives: addAutomaticKeepAlives,
          addRepaintBoundaries: addRepaintBoundaries,
        ),
        super(
          key: key,
          scrollDirection: scrollDirection,
          reverse: reverse,
          controller: controller,
          primary: primary,
          physics: physics,
          shrinkWrap: shrinkWrap,
          padding: padding,
        );

  /// Creates a scrollable, 2D array of widgets of variable sizes with a fixed
  /// number of cells in the cross axis that are created on demand.
  ///
  /// This constructor is appropriate for grid views with a large (or infinite)
  /// number of children because the builder is called only for those children
  /// that are actually visible.
  ///
  /// Uses a [SliverLayoutGridDelegateWithFixedCrossAxisCount] as the
  /// [gridDelegate].
  ///
  ///  Providing a non-null [itemCount] improves the ability of the
  /// [SliverLayoutGridDelegate] to estimate the maximum scroll extent.
  ///
  /// [itemBuilder] will be called only with
  /// indices greater than or equal to
  /// zero and less than [itemCount].
  ///
  /// The `addAutomaticKeepAlives` argument corresponds to the
  /// [SliverLayoutGridDelegate.addAutomaticKeepAlives] property. The
  /// `addRepaintBoundaries` argument corresponds to the
  /// [SliverLayoutGridDelegate.addRepaintBoundaries] property. Both must not be
  /// null.
  LayoutGrid.countBuilder({
    Key? key,
    Axis scrollDirection: Axis.vertical,
    bool reverse: false,
    ScrollController? controller,
    bool? primary,
    ScrollPhysics? physics,
    bool shrinkWrap: false,
    EdgeInsetsGeometry? padding,
    required int crossAxisCount,
    required IndexedWidgetBuilder itemBuilder,
    int? itemCount,
    double mainAxisSpacing: 0.0,
    double crossAxisSpacing: 0.0,
    bool addAutomaticKeepAlives: true,
    bool addRepaintBoundaries: true,
  })  : gridDelegate = SliverLayoutGridDelegateWithFixedCrossAxisCount(
          crossAxisCount: crossAxisCount,
          mainAxisSpacing: mainAxisSpacing,
          crossAxisSpacing: crossAxisSpacing,
        ),
        childrenDelegate = SliverChildBuilderDelegate(
          itemBuilder,
          childCount: itemCount,
          addAutomaticKeepAlives: addAutomaticKeepAlives,
          addRepaintBoundaries: addRepaintBoundaries,
        ),
        super(
          key: key,
          scrollDirection: scrollDirection,
          reverse: reverse,
          controller: controller,
          primary: primary,
          physics: physics,
          shrinkWrap: shrinkWrap,
          padding: padding,
        );

  /// Creates a scrollable, 2D array of widgets of variable sizes with cells
  /// that  each have a maximum cross-axis extent.
  ///
  /// Uses a [SliverGridDelegateWithMaxCrossAxisExtent] as the [gridDelegate].
  ///
  ///  Providing a non-null [itemCount] improves the ability of the
  /// [SliverLayoutGridDelegate] to estimate the maximum scroll extent.
  ///
  /// [itemBuilder] will be called only with
  /// indices greater than or equal to
  /// zero and less than [itemCount].
  ///
  /// The `addAutomaticKeepAlives` argument corresponds to the
  /// [SliverLayoutGridDelegate.addAutomaticKeepAlives] property. The
  /// `addRepaintBoundaries` argument corresponds to the
  /// [SliverLayoutGridDelegate.addRepaintBoundaries] property. Both must not be
  /// null.
  ///
  /// See also:
  ///
  ///  * [new SliverGrid.extent], the equivalent constructor for [SliverGrid].
  LayoutGrid.extent({
    Key? key,
    Axis scrollDirection: Axis.vertical,
    bool reverse: false,
    ScrollController? controller,
    bool? primary,
    ScrollPhysics? physics,
    bool shrinkWrap: false,
    EdgeInsetsGeometry? padding,
    required double maxCrossAxisExtent,
    double mainAxisSpacing: 0.0,
    double crossAxisSpacing: 0.0,
    bool addAutomaticKeepAlives: true,
    bool addRepaintBoundaries: true,
    List<Widget> children: const <Widget>[],
  })  : gridDelegate = SliverLayoutGridDelegateWithMaxCrossAxisExtent(
          maxCrossAxisExtent: maxCrossAxisExtent,
          mainAxisSpacing: mainAxisSpacing,
          crossAxisSpacing: crossAxisSpacing,
        ),
        childrenDelegate = SliverChildListDelegate(
          children,
          addAutomaticKeepAlives: addAutomaticKeepAlives,
          addRepaintBoundaries: addRepaintBoundaries,
        ),
        super(
          key: key,
          scrollDirection: scrollDirection,
          reverse: reverse,
          controller: controller,
          primary: primary,
          physics: physics,
          shrinkWrap: shrinkWrap,
          padding: padding,
        );

  /// Creates a scrollable, 2D array of widgets of variable sizes with cells
  /// that  each have a maximum cross-axis extent that are created on demand.
  ///
  /// This constructor is appropriate for grid views with a large (or infinite)
  /// number of children because the builder is called only for those children
  /// that are actually visible.
  ///
  /// Uses a [SliverGridDelegateWithMaxCrossAxisExtent] as the [gridDelegate].
  ///
  /// The `addAutomaticKeepAlives` argument corresponds to the
  /// [SliverLayoutGridDelegate.addAutomaticKeepAlives] property. The
  /// `addRepaintBoundaries` argument corresponds to the
  /// [SliverLayoutGridDelegate.addRepaintBoundaries] property. Both must not be
  /// null.
  ///
  /// See also:
  ///
  ///  * [new SliverGrid.extent], the equivalent constructor for [SliverGrid].
  LayoutGrid.extentBuilder({
    Key? key,
    Axis scrollDirection: Axis.vertical,
    bool reverse: false,
    ScrollController? controller,
    bool? primary,
    ScrollPhysics? physics,
    bool shrinkWrap: false,
    EdgeInsetsGeometry? padding,
    required double maxCrossAxisExtent,
    required IndexedWidgetBuilder itemBuilder,
    int? itemCount,
    double mainAxisSpacing: 0.0,
    double crossAxisSpacing: 0.0,
    bool addAutomaticKeepAlives: true,
    bool addRepaintBoundaries: true,
  })  : gridDelegate = SliverLayoutGridDelegateWithMaxCrossAxisExtent(
          maxCrossAxisExtent: maxCrossAxisExtent,
          mainAxisSpacing: mainAxisSpacing,
          crossAxisSpacing: crossAxisSpacing,
        ),
        childrenDelegate = SliverChildBuilderDelegate(
          itemBuilder,
          childCount: itemCount,
          addAutomaticKeepAlives: addAutomaticKeepAlives,
          addRepaintBoundaries: addRepaintBoundaries,
        ),
        super(
          key: key,
          scrollDirection: scrollDirection,
          reverse: reverse,
          controller: controller,
          primary: primary,
          physics: physics,
          shrinkWrap: shrinkWrap,
          padding: padding,
        );

  /// A delegate that controls the layout of the children within the
  /// [LayoutGrid].
  ///
  /// The [LayoutGrid] and [LayoutGrid.custom] constructors let you specify this
  /// delegate explicitly. The other constructors create a [gridDelegate]
  /// implicitly.
  final SliverLayoutGridDelegate? gridDelegate;

  /// A delegate that provides the children for the [LayoutGrid].
  ///
  /// The [LayoutGrid.custom] constructor lets you specify this delegate
  /// explicitly. The other constructors create a [childrenDelegate] that wraps
  /// the given child list.
  final SliverChildDelegate childrenDelegate;

  @override
  Widget buildChildLayout(BuildContext context) {
    final layout = Layout.maybeOf(context);
    SliverLayoutGridDelegate defaultDelegate() {
      if (layout != null) {
        return SliverLayoutGridDelegateWithFixedCrossAxisCount(
          crossAxisCount: layout.columns,
          mainAxisSpacing: layout.gutter,
          crossAxisSpacing: layout.gutter,
        );
      } else {
        return SliverLayoutGridDelegateWithFixedCrossAxisCount(
          crossAxisCount: 12,
          mainAxisSpacing: 0,
          crossAxisSpacing: 0,
        );
      }
    }

    return SliverLayoutGrid(
      delegate: childrenDelegate,
      gridDelegate: gridDelegate ?? defaultDelegate(),
    );
  }
}
