import 'package:layout/layout.dart';

import '../rendering/sliver_layout_grid.dart';
import 'sliver.dart';

import 'package:flutter/rendering.dart';
import 'package:flutter/widgets.dart';

class SliverLayoutGrid extends StatelessWidget {
  const SliverLayoutGrid({
    Key? key,
    required this.delegate,
    this.gridDelegate,
  });

  /// Creates a sliver that places multiple box children in a two dimensional
  /// arrangement with a fixed number of cells in the cross axis.
  ///
  /// Uses a [SliverLayoutGridDelegateWithFixedCrossAxisCount] as the [gridDelegate],
  /// and a [SliverLayoutGridDelegate] as the [delegate].
  ///
  /// See also:
  ///
  ///  * [new LayoutGrid.count], the equivalent constructor for [LayoutGrid] widgets.
  SliverLayoutGrid.count({
    Key? key,
    required int crossAxisCount,
    double mainAxisSpacing: 0.0,
    double crossAxisSpacing: 0.0,
    List<Widget> children: const <Widget>[],
  })  : gridDelegate = SliverLayoutGridDelegateWithFixedCrossAxisCount(
          crossAxisCount: crossAxisCount,
          mainAxisSpacing: mainAxisSpacing,
          crossAxisSpacing: crossAxisSpacing,
        ),
        delegate = SliverChildListDelegate(
          children,
          addAutomaticKeepAlives: true,
          addRepaintBoundaries: true,
          addSemanticIndexes: false,
        ),
        super(key: key);

  /// Creates a sliver that builds multiple box children in a two dimensional
  /// arrangement with a fixed number of cells in the cross axis.
  ///
  /// This constructor is appropriate for grid views with a large (or infinite)
  /// number of children because the builder is called only for those children
  /// that are actually visible.
  ///
  /// Uses a [SliverLayoutGridDelegateWithFixedCrossAxisCount] as the
  /// [gridDelegate], and a [SliverVariableSizeChildBuilderDelegate] as the [delegate].
  ///
  /// See also:
  ///
  ///  * [new LayoutGrid.countBuilder], the equivalent constructor for
  ///  [LayoutGrid] widgets.
  SliverLayoutGrid.countBuilder({
    Key? key,
    required int crossAxisCount,
    required IndexedWidgetBuilder itemBuilder,
    required int itemCount,
    double mainAxisSpacing: 0.0,
    double crossAxisSpacing: 0.0,
  })  : gridDelegate = SliverLayoutGridDelegateWithFixedCrossAxisCount(
          crossAxisCount: crossAxisCount,
          mainAxisSpacing: mainAxisSpacing,
          crossAxisSpacing: crossAxisSpacing,
        ),
        delegate = SliverChildBuilderDelegate(
          itemBuilder,
          childCount: itemCount,
          addAutomaticKeepAlives: false,
          addRepaintBoundaries: false,
          addSemanticIndexes: false,
        ),
        super(key: key);

  /// Creates a sliver that places multiple box children in a two dimensional
  /// arrangement with cells that each have a maximum cross-axis extent.
  ///
  /// Uses a [SliverLayoutGridDelegateWithMaxCrossAxisExtent] as the [gridDelegate],
  /// and a [SliverLayoutGridDelegate] as the [delegate].
  ///
  /// See also:
  ///
  ///  * [new LayoutGrid.extent], the equivalent constructor for [LayoutGrid] widgets.
  SliverLayoutGrid.extent({
    Key? key,
    required double maxCrossAxisExtent,
    double mainAxisSpacing: 0.0,
    double crossAxisSpacing: 0.0,
    List<Widget> children: const <Widget>[],
  })  : gridDelegate = SliverLayoutGridDelegateWithMaxCrossAxisExtent(
          maxCrossAxisExtent: maxCrossAxisExtent,
          mainAxisSpacing: mainAxisSpacing,
          crossAxisSpacing: crossAxisSpacing,
        ),
        delegate = SliverChildListDelegate(
          children,
          addAutomaticKeepAlives: false,
          addRepaintBoundaries: false,
          addSemanticIndexes: false,
        ),
        super(key: key);

  /// Creates a sliver that builds multiple box children in a two dimensional
  /// arrangement with cells that each have a maximum cross-axis extent.
  ///
  /// This constructor is appropriate for grid views with a large (or infinite)
  /// number of children because the builder is called only for those children
  /// that are actually visible.
  ///
  /// Uses a [SliverLayoutGridDelegateWithMaxCrossAxisExtent] as the
  /// [gridDelegate], and a [SliverVariableSizeChildBuilderDelegate] as the [delegate].
  ///
  /// See also:
  ///
  ///  * [new LayoutGrid.extentBuilder], the equivalent constructor for
  ///  [LayoutGrid] widgets.
  SliverLayoutGrid.extentBuilder({
    Key? key,
    required double maxCrossAxisExtent,
    required IndexedWidgetBuilder itemBuilder,
    required int itemCount,
    double mainAxisSpacing: 0.0,
    double crossAxisSpacing: 0.0,
  })  : gridDelegate = SliverLayoutGridDelegateWithMaxCrossAxisExtent(
          maxCrossAxisExtent: maxCrossAxisExtent,
          mainAxisSpacing: mainAxisSpacing,
          crossAxisSpacing: crossAxisSpacing,
        ),
        delegate = SliverChildBuilderDelegate(
          itemBuilder,
          childCount: itemCount,
          addAutomaticKeepAlives: false,
          addRepaintBoundaries: false,
          addSemanticIndexes: false,
        ),
        super(
          key: key,
        );

  /// The delegate that controls the size and position of the children.
  final SliverLayoutGridDelegate? gridDelegate;

  final SliverChildDelegate delegate;

  @override
  Widget build(BuildContext context) {
    final layout = Layout.maybeOf(context);
    SliverLayoutGridDelegate defaultDelegate() {
      if (layout != null) {
        return SliverLayoutGridDelegateWithFixedCrossAxisCount(
          crossAxisCount: layout.columns,
          mainAxisSpacing: layout.gutter,
          crossAxisSpacing: layout.gutter,
        );
      } else {
        return SliverLayoutGridDelegateWithFixedCrossAxisCount(
          crossAxisCount: 12,
          mainAxisSpacing: 0,
          crossAxisSpacing: 0,
        );
      }
    }

    return _RawSliverLayoutGrid(
      delegate: delegate,
      gridDelegate: gridDelegate ?? defaultDelegate(),
    );
  }
}

class _RawSliverLayoutGrid extends SliverVariableSizeBoxAdaptorWidget {
  /// Creates a sliver that places multiple box children in a two dimensional
  /// arrangement.
  const _RawSliverLayoutGrid({
    Key? key,
    required SliverChildDelegate delegate,
    required this.gridDelegate,
  }) : super(key: key, delegate: delegate);

  /// The delegate that controls the size and position of the children.
  final SliverLayoutGridDelegate gridDelegate;

  @override
  RenderSliverLayoutGrid createRenderObject(BuildContext context) {
    final SliverVariableSizeBoxAdaptorElement element =
        context as SliverVariableSizeBoxAdaptorElement;
    return RenderSliverLayoutGrid(
      childManager: element,
      gridDelegate: gridDelegate,
    );
  }

  @override
  void updateRenderObject(
      BuildContext context, RenderSliverLayoutGrid renderObject) {
    renderObject.gridDelegate = gridDelegate;
  }

  @override
  double? estimateMaxScrollOffset(
    SliverConstraints constraints,
    int firstIndex,
    int lastIndex,
    double leadingScrollOffset,
    double trailingScrollOffset,
  ) {
    return super.estimateMaxScrollOffset(
      constraints,
      firstIndex,
      lastIndex,
      leadingScrollOffset,
      trailingScrollOffset,
    );
  }
}
