import 'package:flutter/material.dart';
import 'package:flutter_test/flutter_test.dart';

import 'package:layout_grid/layout_grid.dart';

// Default test screen size defined by Flutter
const kScreenSize = const Size(800, 600);

void main() {
  testWidgets('defaultTestSize', (WidgetTester tester) async {
    final testSize = tester.binding.window.physicalSize /
        tester.binding.window.devicePixelRatio;
    expect(testSize, equals(kScreenSize));
  });
  testWidgets('Empty Grid don`t crash', (WidgetTester tester) async {
    await tester.pumpWidget(RootWidget(
      child: LayoutGrid(),
    ));
    final exception = tester.takeException();
    expect(exception, isNull);
  });
  testWidgets('Empty SliverGrid don`t crash', (WidgetTester tester) async {
    await tester.pumpWidget(RootSliverWidget(
      sliver: SliverLayoutGrid(
        delegate: SliverChildListDelegate([]),
      ),
    ));
    final exception = tester.takeException();
    expect(exception, isNull);
  });

  testWidgets(
    'Default Grid has 12 columns a 0 spacing',
    (WidgetTester tester) async {
      final size = 60.0;
      await tester.pumpWidget(RootWidget(
        width: 12 * size, // 12x60 width
        child: LayoutGrid(
          children: [
            for (int i = 0; i < 12; i++)
              Cell(
                snap: 1,
                key: ValueKey('$i'),
                extent: size,
                child: Container(),
              ),
          ],
        ),
      ));

      for (int i = 0; i < 12; i++) {
        final cellFinder = find.byKey(ValueKey('$i'));
        expect(cellFinder, findsOneWidget);
        final rect = tester.getRect(cellFinder);
        expect(rect.top, equals(0));
        expect(rect.left, equals(size * i));
        expect(rect.width, equals(size));
        expect(rect.height, equals(size));
      }
    },
  );

  testWidgets(
    'Default SliverGrid has 12 columns a 0 spacing',
    (WidgetTester tester) async {
      final size = 60.0;
      await tester.pumpWidget(RootSliverWidget(
        width: 12 * size, // 12x60 width
        sliver: SliverLayoutGrid(
          delegate: SliverChildListDelegate(
            [
              for (int i = 0; i < 12; i++)
                Cell(
                  snap: 1,
                  key: ValueKey('$i'),
                  extent: size,
                  child: Container(),
                ),
            ],
          ),
        ),
      ));

      for (int i = 0; i < 12; i++) {
        final cellFinder = find.byKey(ValueKey('$i'));
        expect(cellFinder, findsOneWidget);
        final rect = tester.getRect(cellFinder);
        expect(rect.top, 0);
        expect(rect.left, size * i);
        expect(rect.width, size);
        expect(rect.height, size);
      }
    },
  );

  testWidgets(
    'Custom columns and spacing',
    (WidgetTester tester) async {
      await tester.pumpWidget(RootWidget(
        width: 690, // 10 col x 60 width +  9 spaces x 10 width
        child: LayoutGrid(
          gridDelegate: SliverLayoutGridDelegateWithFixedCrossAxisCount(
            crossAxisCount: 10,
            mainAxisSpacing: 20,
            crossAxisSpacing: 10,
          ),
          children: [
            for (int i = 0; i < 20; i++)
              Cell(
                snap: 1,
                key: ValueKey('$i'),
                extent: 60,
                child: Container(),
              ),
          ],
        ),
      ));

      for (int i = 0; i < 10; i++) {
        final cellFinder = find.byKey(ValueKey('$i'));
        expect(cellFinder, findsOneWidget);
        expect(
          tester.getRect(cellFinder),
          Rect.fromLTWH((60.0 + 10) * i, 0, 60, 60),
        );
      }
      // second line
      for (int i = 0; i < 10; i++) {
        final cellFinder = find.byKey(ValueKey('${i + 10}'));
        expect(cellFinder, findsOneWidget);
        expect(
          tester.getRect(cellFinder),
          Rect.fromLTWH((60.0 + 10) * i, 60 + 20, 60, 60),
        );
      }
    },
  );

  testWidgets(
    'Normal Grid columns and spacing',
    (WidgetTester tester) async {
      await tester.pumpWidget(RootWidget(
        width: 600, // 10 col x 60 width +  9 spaces x 10 width
        child: LayoutGrid(
          gridDelegate: SliverLayoutGridDelegateWithFixedCrossAxisCount(
            crossAxisCount: 10,
            masonry: false,
          ),
          children: [
            for (int i = 0; i < 10; i++)
              Cell(
                snap: 1,
                key: ValueKey('$i'),
                extent: 10 * (i + 1),
                child: Container(),
              ),
            for (int i = 0; i < 10; i++)
              Cell(
                snap: 1,
                key: ValueKey('${10 + i}'),
                extent: 100 - 10.0 * i,
                child: Container(),
              ),
          ],
        ),
      ));

      for (int i = 0; i < 10; i++) {
        final cellFinder = find.byKey(ValueKey('$i'));
        expect(cellFinder, findsOneWidget);
        expect(
          tester.getRect(cellFinder),
          Rect.fromLTWH(60.0 * i, 0, 60, 10 * (i + 1)),
        );
      }
      // second line
      for (int i = 0; i < 10; i++) {
        final cellFinder = find.byKey(ValueKey('${i + 10}'));
        expect(cellFinder, findsOneWidget);
        expect(
          tester.getRect(cellFinder),
          Rect.fromLTWH(60.0 * i, 100, 60, 100 - 10.0 * i),
        );
      }
    },
  );

  testWidgets(
    'Manonry Grid columns and spacing',
    (WidgetTester tester) async {
      await tester.pumpWidget(RootWidget(
        width: 600, // 10 col x 60 width +  9 spaces x 10 width
        child: LayoutGrid(
          gridDelegate: SliverLayoutGridDelegateWithFixedCrossAxisCount(
            crossAxisCount: 10,
            masonry: true,
          ),
          children: [
            for (int i = 0; i < 10; i++)
              Cell(
                snap: 1,
                key: ValueKey('$i'),
                extent: 10 * (i + 1),
                child: Container(),
              ),
            for (int i = 0; i < 10; i++)
              Cell(
                snap: 1,
                key: ValueKey('${10 + i}'),
                extent: 100 - 10.0 * i,
                child: Container(),
              ),
          ],
        ),
      ));

      for (int i = 0; i < 10; i++) {
        final cellFinder = find.byKey(ValueKey('$i'));
        expect(cellFinder, findsOneWidget);
        expect(
          tester.getRect(cellFinder),
          Rect.fromLTWH(60.0 * i, 0, 60, 10 * (i + 1)),
        );
      }
      // second line
      for (int i = 0; i < 10; i++) {
        final cellFinder = find.byKey(ValueKey('${i + 10}'));
        expect(cellFinder, findsOneWidget);
        expect(
          tester.getRect(cellFinder),
          Rect.fromLTWH(60.0 * i, 10 * (i + 1), 60, 100 - 10.0 * i),
        );
      }
    },
  );
}

class RootWidget extends StatelessWidget {
  final double? width;
  final Widget child;

  const RootWidget({
    Key? key,
    this.width,
    required this.child,
  }) : super(key: key);

  @override
  Widget build(BuildContext context) {
    return MaterialApp(
      home: Container(
        alignment: Alignment.topLeft,
        child: Container(
          width: width,
          child: child,
        ),
      ),
    );
  }
}

class RootSliverWidget extends StatelessWidget {
  final double? width;
  final Widget sliver;

  const RootSliverWidget({
    Key? key,
    this.width,
    required this.sliver,
  }) : super(key: key);

  @override
  Widget build(BuildContext context) {
    return RootWidget(
      width: width,
      child: CustomScrollView(
        slivers: [
          sliver,
        ],
      ),
    );
  }
}
